import re
import sys
import imp
import a99
from f311 import FilePy

__all__ = ["FileOptions"]


@a99.froze_it
class FileOptions(FilePy):
    """
    PFANT Command-line Options

    For each xxxx attribute not starting with "_" there exists
    a variable in config.f90 named config_xxxx, and
    a command-line option "--xxxx"

    """

    default_filename = "options.py"
    attrs = []
    editors = ["x.py"]

    def __init__(self):
        FilePy.__init__(self)

        # innewmarcs, hydro2, pfant, nulbad
        self.logging_level = None
        self.logging_console = None
        self.logging_file = None
        self.fn_logging = None
        self.fn_main = None
        self.explain = None
        self.play = None

        # innewmarcs, hydro2, pfant
        self.fn_modeles = None

        # innewmarcs
        self.fn_modgrid = None
        self.fn_moo = None
        self.allow = None

        # innewmarcs, pfant
        self.opa = None
        self.fn_opa = None

        # hydro2, pfant
        self.fn_absoru2 = None
        self.fn_hmap = None
        self.llzero = None
        self.llfin = None

        # hydro2
        self.zph = None
        self.kik = None
        self.amores = None
        self.kq = None

        # pfant
        self.fn_dissoc    = None
        self.fn_partit    = None
        self.fn_abonds    = None
        self.fn_atoms   = None
        self.fn_molecules = None
        self.fn_mollist = None
        self.fn_lines     = None
        self.fn_log       = None
        self.fn_progress = None
        self.flprefix = None
        self.no_molecules = None
        self.no_h = None
        self.no_atoms = None
        self.zinf = None
        self.pas = None
        self.aint = None
        self.interp = None
        self.abs = None
        self.sca = None
        self.absoru = None

        # nulbad
        self.norm = None
        self.flam = None
        self.convol = None
        self.fwhm = None
        self.pat = None
        self.fn_flux = None
        self.fn_cv = None


    def init_default(self):
        """Overriden to reset all options to None."""
        for option_name in self.get_names():
            self.__setattr__(option_name, None)

    def get_names(self):
        """Returns a list with the names of all the options. Names come sorted"""
        d = dir(FilePy())
        return [x for x in dir(self) if not x.startswith(('_', 'get_')) and
         x not in d]

    def get_args(self):
        """
        Returns a list of command-line arguments (only options that have been set)
        """
        l = []
        names = self.get_names()
        for attr_name in names:
            value = self.__getattribute__(attr_name)
            if value is not None:
                s_value = ("T" if value else "F") if isinstance(value, bool) else str(value)
                if re.search(r"[,|& ]", s_value):
                    s_value = '"'+s_value+'"'  # adds quotes if string contains one of the characters above
                l.extend(["--"+attr_name, s_value])
        return l

    def _do_load(self, filename):
        with open(filename, "r") as h:
            lines = h.read()
        cfg = imp.new_module('cfg')
        exec(lines, cfg.__dict__)

        valid_names = self.get_names()
        for option_name in valid_names:
            # resets all options, because if they are not specified in the file, they should be not set
            self.__setattr__(option_name, None)

        for option_name in dir(cfg):
            if option_name.startswith("_"):
                continue
            if not option_name in valid_names:
                raise RuntimeError("Invalid option: '%s'" % option_name)
            self.__setattr__(option_name, cfg.__dict__[option_name])

    def _do_save_as(self, filename):
        l = []
        names = self.get_names()
        for attr_name in names:
            value = self.__getattribute__(attr_name)
            if value is not None:
                l.append("%s = %s" % (attr_name, repr(value)))

        with open(filename, "w") as h:
            h.write(self._get_magic()+"\n#\n")
            h.write("# Command-line options\n")
            h.write("# Note: do not edit this file directly, as any changes will be overwritten by x.py\n\n")

            h.write(("\n".join(l))+"\n")

_options = FileOptions()
FileOptions.attrs = _options.get_names()


from .a_WContinua import *
from .a_WContinuum import *
from .a_WDBFCF import *
from .a_WDBMolecule import *
from .a_WDBPFANTMol import *
from .a_WDBState import *
from .a_WDBSystem import *
from .a_WDBSystemPfantmolFCF import *
from .a_WFileAbonds import *
from .a_WFileAbXFwhm import *
from .a_WFileMain import *
from .a_WFileMolConsts import *
from .a_WFileMolDB import *
from .a_WMolecularConstants import *
from .a_WOptionsEditor import *
from .a_XAtomLinesEditor import *
from .a_XConvMol import *
from .a_XFileAbonds import *
from .a_XFileAtoms import *
from .a_XFileAtomsHistogram import *
from .a_XFileMain import *
from .a_XFileMolConsts import *
from .a_XFileMolDB import *
from .a_XFileMolecules import *
from .a_XFileOptions import *
from .a_XMainAbonds import *
from .a_XMolLinesEditor import *
from .a_XPFANT import *
from .a_XRunnableManager import *
from .a_XMulti import *
from .a_XFileMoleculeHistogram import *
from . import _shared
# from .a_WFileAbonds import *
# from .a_WFileAbXFwhm import *
# from .a_WFileMain import *
# from .a_WOptionsEditor import *
# from .a_XAtomLinesEditor import *
# from .a_XFileAbonds import *
# from .a_XFileAtoms import *
# from .a_XFileAtomsHistogram import *
# from .a_XFileMain import *
# from .a_XFileMolecules import *
# from .a_XMainAbonds import *
# from .a_XMolLinesEditor import *
# from .a_WContinuum import *
# from .a_WContinua import *
# from .a_XFileOptions import *
#
# from . import a_XMulti
# from . import a_XPFANT
# from . import a_XRunnableManager
# from . import _shared
#
# from .a_XMulti import *
# from .a_XPFANT import *
# from .a_XRunnableManager import *
#
# from .a_XFileMolDB import *
# from .a_XFileMolConsts import *
# from .a_WMolecularConstants import *
# from .a_XConvMol import *



#!/usr/bin/env python

"""
Editor for molecules SQLite database

This application can edit files of class FileMolDB.
"""

import pyfant
import sys
import argparse
import a99
import logging
import os


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":

    deffn = pyfant.FileMolDB.default_filename

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='Molecules database file name',
                        default=deffn, nargs='?')
    args = parser.parse_args()

    if args.fn == deffn and not os.path.isfile(deffn):
        args.fn = None

    m = None
    if args.fn is not None:
        m = pyfant.FileMolDB()
        m.load(args.fn)
    app = a99.get_QApplication([])
    form = pyfant.XFileMolDB(None, [m,])
    form.showMaximized()
    sys.exit(app.exec_())










"""
VALD3-specific conversion
"""


import a99
# from .calc_qgbd import calc_qgbd_tio_like
from .convlog import *
from collections import OrderedDict


__all__ = ["vald3_to_sols"]



def vald3_to_sols(molconsts, file_vald3, qgbd_calculator):
    """
    Converts VALD3 molecular lines data to PFANT "sets of lines"

    Args:
        molconsts: a dict-like object combining field values from tables 'molecule', 'state',
                    and 'pfantmol' from a FileMolDB database
        file_vald3: FileVald3 instance with only one species
        qgbd_calculator: callable that can calculate "qv", "gv", "bv", "dv",
                         e.g., calc_qbdg_tio_like()

    Returns: (a list of ftpyfant.SetOfLines objects, a MolConversionLog object)
    """

    def append_error(msg):
        log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

    # C     BAND (v',v'')=(VL,V2L)
    # C     WN: vacuum wavenumber   WL : air wavelength
    # C     J2L: lower state angular momentum number
    # C     iso: isotope/ 26: 12C16O, 36: 13C16O, 27: 12C17O, 28: 12C18O
    # C     ramo: branch, for P: ID=1, for R: ID=2
    # C     name: file name/ coisovLv2L
    # C     HL: Honl-London factor
    # C     FR: oscillator strength

    if not isinstance(file_vald3, pyfant.FileVald3):
        raise TypeError("Invalid type for argument 'file_vald3': {}".format(type(file_vald3)))
    if len(file_vald3) > 1:
        raise ValueError("Argument 'file_vald3' must have only one species, but it has {}".format(len(file_vald3)))

    lines = file_vald3.speciess[0].lines
    n = len(lines)

    S = molconsts.get_S2l()
    DELTAK = molconsts.get_deltak()

    sols = OrderedDict()  # one item per (vl, v2l) pair
    log = MolConversionLog(n)

    for i, line in enumerate(lines):
        assert isinstance(line, pyfant.Vald3Line)
        try:
            wl = line.lambda_
            # Br, J2l = f_group(data["local_lower_quanta"][i])
            # Jl = _Jl(Br, J2l)
            # V = f_class(data["global_upper_quanta"][i])
            # V_ = f_class(data["global_lower_quanta"][i])
            # A = data["a"][i]

            # A seguir normalizacion de factor de Honl-london (HLN)
            Normaliza = 1/((2.0*line.J2l+1)*(2.0*S+1)*(2.0-DELTAK))

            # A seguir teremos a forca de oscilador
            # mas q sera normalizada segundo o programa da Beatriz
            # gf = Normaliza*1.499*(2*Jl+1)*A/(nu**2)  <-- this was for HITRAN
            gf_pfant = Normaliza*10**line.loggf

            J2l_pfant = int(line.J2l)  # ojo, estamos colocando J2L-0.5!
        except Exception as e:
            log.errors.append("#{}{} line: {}".format(i+1, a99.ordinal_suffix(i+1), str(e)))
            continue

        sol_key = "%3d%3d" % (line.vl, line.v2l)  # (v', v'') transition (v_sup, v_inf)
        if sol_key not in sols:
            qgbd = qgbd_calculator(molconsts, line.v2l)
            qqv = qgbd["qv"]
            ggv = qgbd["gv"]
            bbv = qgbd["bv"]
            ddv = qgbd["dv"]
            sols[sol_key] = pyfant.SetOfLines(line.vl, line.v2l, qqv, ggv, bbv, ddv, 1.)

        sol = sols[sol_key]
        # TODO assuming singlet!!!
        sol.append_line(wl, gf_pfant, J2l_pfant, ph.singlet.quanta_to_branch(line.Jl, line.J2l))

    return (list(sols.values()), log)



# """
# Transition-specific calculations
# 
# References:
#   [1] agrup.plez7.f
# """
# 
# from collections import OrderedDict
# 
# 
# __all__ = ["calc_qgbd_tio_like"]
# 
# 
# def calc_qgbd_tio_like(molconsts, v_lo):
#     """
#     Calculates qv, gv, bv, dv in TiO-like fashion
# 
#     Based on Fortran source 'agrup.plez7.f'
# 
#     Args:
#         molconsts: dict-like object with keys:
#             "omega_e"
#             "omega_ex_e"
#             "omega_ey_e"
#             "B_e"
#             "alpha_e"
#             "D_e"
#             "beta_e"
# 
#         v_lo: (integer) low transition level
# 
#     Returns: qbdg, i.e., a dictionary with keys:
#             "qv", "gv", "bv", "dv", "gzero"
#     """
# 
#     omega_e = molconsts["state2l_omega_e"]
#     omega_ex_e = molconsts["state2l_omega_ex_e"]
#     omega_ey_e = molconsts["state2l_omega_ey_e"]
#     B_e = molconsts["state2l_B_e"]
#     D_e = molconsts["state2l_D_e"]
#     alpha_e = molconsts["state2l_alpha_e"]
#     beta_e = molconsts["state2l_beta_e"]
# 
#     if int(v_lo) != v_lo:
#         raise ValueError("Argument 'v_lo' must be an integer")
# 
#     gzero = omega_e / 2.0 - omega_ex_e / 4.0 + omega_ey_e / 8.0
#     v_lo5 = v_lo + 0.5
# 
#     # Franck-Condon factor. Currently always incorporated into "SJ"
#     qv = 1.
#     # rotational constant "B_v"
#     bv = B_e - alpha_e * v_lo5
#     # rotational constant "D_v"
#     dv = (D_e + beta_e * v_lo5) * 1.0e+06
#     # rotational term " G_A"
#     gv = omega_e * v_lo5 - omega_ex_e * v_lo5 ** 2 + omega_ey_e * v_lo5 ** 3 - gzero
# 
#     qgbd = OrderedDict((("qv", qv), ("gv", gv), ("bv", bv), ("dv", dv), ("gzero", gzero)))
# 
#     return qgbd



"""
Visualization classes for atmospheric models
"""

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D  # yes, required (see below)
import pyfant
from f311 import Vis


__all__ = [
"VisMarcs", "VisOpa", "VisModCurves", "VisMarcsSaveAsMod", "VisVector", "VisGrid", "VisModRecord", ]

class VisModRecord(Vis):
    """
    Plots vectors (nh, teta, pe, pg, log_tau_ross) of ".mod" file in 5 subplots sharing
    the same x-axis.
    """

    input_classes = (pyfant.FileModBin,)
    action = "Visualize single record"

    def __init__(self):
        Vis.__init__(self)
        # 1-based (first is 1 (as in Fortran), not 0) record index in .mod file
        self.inum = None

    def _do_use(self, obj):
        assert isinstance(obj, pyfant.FileModBin)
        n = len(obj.records)

        if n == 1 and self.inum is None:
            self.inum = 1
        if n > 1 and self.inum is None:
            inum, ok = QInputDialog.getInt(None, "Record number",
             "Enter record number (1 to {0:d})".format(n), 1, 1, n)
            if ok:
                self.inum = inum

        if self.inum is not None:
            r = obj.records[self.inum-1]
            _plot_mod_record("{0!s} - {1!s}".format(self.title, repr(r)), r)


class VisMarcs(Vis):
    """
    Similar to VisModRecord but accepts FileModTxt
    """

    input_classes = (pyfant.FileModTxt,)
    action = "Visualize model"

    def __init__(self):
        Vis.__init__(self)

    def _do_use(self, obj):
        _plot_mod_record(self.title, obj.record)


class VisMarcsSaveAsMod(Vis):
    """
    Asks user for file name and saves as a binary .mod file
    """

    input_classes = (pyfant.FileModTxt,)
    action = 'Save as a binary ".mod" file'

    def __init__(self):
        Vis.__init__(self)

    def _do_use(self, obj):
        d = "."  # todo find a way to pass current directory in a_Xexplorer (not pwd)
        new_filename = QFileDialog.getSaveFileName(None,
         self.action.capitalize(), d, "*.mod")[0]
        if new_filename:
            f = pyfant.FileModBin()
            f.records = [obj.record]
            f.save_as(str(new_filename))
        return False


def _plot_mod_record(title, r):
    f, axarr = plt.subplots(5, sharex=True)
    f.canvas.set_window_title(title)
    x = np.linspace(1, r.ntot, r.ntot)

    axarr[0].plot(x, r.nh)
    axarr[0].set_ylabel('nh')
    axarr[1].plot(x, r.teta)
    axarr[1].set_ylabel('teta')
    axarr[2].plot(x, r.pe)
    axarr[2].set_ylabel('pe')
    axarr[3].plot(x, r.pg)
    axarr[3].set_ylabel('pg')
    axarr[4].plot(x, r.log_tau_ross)
    axarr[4].set_ylabel('log_tau_ross')
    axarr[4].set_xlabel("Atmospheric layer #")
    for i in range(5):
        ax = axarr[i]
        ax.set_xlim([.5, r.ntot+.5])
    plt.tight_layout()
    plt.show()


class VisModCurves(Vis):
    """
    Plots vectors
    (teff, glog, asalog, asalalf, nhe) in 2D (record #)x(value) plots, and
    (nh, teta, pe, pg, log_tau_ross) (layer #)x(record #)x(value) 3D plots
    """

    input_classes = (pyfant.FileMoo, pyfant.FileModBin)
    action = "(nh, teta, pe, pg, log_tau_ross) per layer curves in 3D"

    def _do_use(self, m):
        nr = len(m)

        #################
        # 3D plots
        vars = ['nh', 'teta', 'pe', 'pg', 'log_tau_ross']
        for var in vars:

            fig = plt.figure()
            if self.parent_form:
                fig.canvas.setParent(self.parent_form)
            ax = fig.gca(projection='3d')
            fig.canvas.set_window_title('{0!s} -- {1!s}'.format(self.title, var))
            rr = m.records

            for i, r in enumerate(rr):
                x = np.linspace(1, r.ntot, r.ntot)
                y = np.ones(len(x)) * (i + 1)
                z = r.__getattribute__(var)
                ax.plot(x, y, z, label='a', color='k')

            ax.set_xlabel('Atmospheric layer #')
            ax.set_ylabel('Record number')
            ax.set_zlabel(var)

        plt.show()


class VisGrid(Vis):
    __doc__ = """(glog, teff, [Fe/H]) 3D scatterplot"""

    input_classes = (pyfant.FileMoo, pyfant.FileModBin)
    action = __doc__

    def _do_use(self, m):
        asalog, teff, glog = [], [], []
        for r in m.records:
            asalog.append(r.asalog)
            teff.append(r.teff)
            glog.append(r.glog)

        # teff-glog-asalog scatterplot
        fig = plt.figure()
        ax = fig.gca(projection='3d')
        ax.scatter(asalog, teff, glog, c='r', s=60, marker='o')
        ax.set_xlabel('asalog ([Fe/H] relative to Sun)')
        ax.set_ylabel('teff')
        ax.set_zlabel('glog')
        fig.canvas.set_window_title(self.title+" -- asalog-teff-glog scatterplot")
        plt.tight_layout()
        plt.show()


class VisVector(Vis):
    __doc__ = """(glog, teff, [Fe/H]) same-x-axis stacked subplots"""

    input_classes = (pyfant.FileMoo, pyfant.FileModBin)
    action = __doc__

    def _do_use(self, m):
        asalog, teff, glog = [], [], []
        for r in m.records:
            asalog.append(r.asalog)
            teff.append(r.teff)
            glog.append(r.glog)

        # 3 subplots sharing same x-axis
        ll = ['asalog', 'teff', 'glog']
        f, axarr = plt.subplots(len(ll), sharex=True)
        nr = len(asalog)
        x = np.linspace(1, nr, nr)
        rr = m.records
        for i, a in enumerate(ll):
            v = eval(a)
            axarr[i].plot(x, v)
            axarr[i].set_ylabel(a)
        axarr[len(ll)-1].set_xlabel("Record #")
        f.canvas.set_window_title("{0!s} -- {1!s}".format(self.title, 'one-value-per-model'))
        plt.tight_layout()

        plt.show()


class VisOpa(Vis):
    """
    Visualizer for FileOpa class

    Plots vectors ???
    """

    input_classes = (pyfant.FileOpa,)
    action = "Visualize opacities file"

    def _do_use(self, obj):
        assert isinstance(obj, pyfant.FileOpa)

        # 8 subplots sharing same x-axis
        ll = ["rad", "tau", "t", "pe", "pg", "rho", "xi", "ops"]
        titles = ["spherical radiative transfer",
                  "continuumm optical depth at {0:g} angstrom".format(obj.swave),
                  "temperature (K)",
                  "electron pressure (dyn/cm**2)",
                  "total gas pressure (dyn/cm**2)",
                  "densigy (g/cm**3)",
                  "microturbulence parameter (km/s)",
                  "continuumm opacity at {0:g} angstrom (cm**2/g)".format(obj.swave)]

        f, axarr = plt.subplots(nrows=4, ncols=2, sharex=True)
        x = np.linspace(1, obj.ndp, obj.ndp)
        i = 0
        for m in range(4):
            for n in range(2):
                a = ll[i]
                ax = axarr[m, n]
                ax.plot(x, obj.__getattribute__(a))
                ax.set_title("{0!s}: {1!s}".format(a, titles[i]))

                i += 1
        axarr[3, 0].set_xlabel("Layer #")
        axarr[3, 1].set_xlabel("Layer #")

        f.canvas.set_window_title("{0!s} -- {1!s}".format(self.title, 'one-value-per-model'))
        plt.tight_layout()

        #################
        # 3D plots
        vars = ["abs", "sca"]
        titles = ["specific continuous absorption opacity (cm**2/g)",
                  "specific continuous scattering opacity (cm**2/g)"]

        for var, title in zip(vars, titles):
            attr = obj.__getattribute__(var)
            x = np.log10(obj.wav)

            fig = plt.figure()
            if self.parent_form:
                fig.canvas.setParent(self.parent_form)
            ax = fig.gca(projection='3d')
            fig.canvas.set_window_title('{0!s} -- {1!s}'.format(self.title, var))

            for k in range(obj.ndp):
                y = np.ones(len(x)) * (k + 1)
                # z = attr[:, k]
                z = np.log10(attr[:, k])
                ax.plot(x, y, z, label='a', color='k')
                # ax.set_xscale("log")
                # ax.semilogx(x, y, z, label='a', color='k')

            ax.set_xlabel('log10(wavelength)')
            ax.set_ylabel('Layer #')
            ax.set_zlabel('log10({0!s})'.format(var))
            ax.set_title("{0!s}: {1!s}".format(var, title))

        plt.show()




"""
Miscellanea routines that depend on other pyfant modules.

Rule: no pyfant module can import util!!!

"""
import copy
import os
import a99
import pyfant
from . import runnables

__all__ = ["MultiRunnable"]

__multi_id_maker = None

def _multi_id_maker():
    global __multi_id_maker
    if __multi_id_maker is None:
        import pyfant as pf
        __multi_id_maker = pyfant.IdMaker()
        __multi_id_maker.session_prefix_singular = pyfant.MULTISESSION_PREFIX
    return __multi_id_maker


@a99.froze_it
class MultiRunnableStatus(object):
    def __init__(self, runnable):
        assert isinstance(runnable, MultiRunnable)
        self.runnable = runnable
        self.stage = ""

    def __str__(self):
        l = []
        if self.runnable.flag_finished:
            l.append("finished")
        if self.runnable.flag_running:
            l.append("running - "+self.stage)
        if self.runnable.flag_killed:
            l.append("*killed*")
        if self.runnable.flag_error:
            l.append("*error*")
        if self.runnable.error_message:
            l.append("*" + str(self.runnable.error_message) + "*")
        if len(l) > 0:
            return " ".join(l)
        return "?"


@a99.froze_it
class MultiRunnable(runnables.Runnable):
    """
    Differential abundances X FWHM's runnable.
    """

    def __init__(self, file_main, file_abonds, options, file_abxfwhm, custom_id=None):
        import pyfant as pf
        pyfant.Runnable.__init__(self)
        assert isinstance(file_main, pyfant.FileMain)
        assert isinstance(file_abonds, pyfant.FileAbonds)
        assert isinstance(options, pyfant.FileOptions)
        assert isinstance(file_abxfwhm, pyfant.FileAbXFwhm)
        self.__file_main = file_main
        self.__file_abonds = file_abonds
        self.__options = options
        self.__file_abxfwhm = file_abxfwhm
        self.__custom_id = custom_id

        # # Protected variables
        # ExecutableStatus instance
        self.__status = MultiRunnableStatus(self)

        # # Private variables
        self.__logger = a99.get_python_logger()
        self.__sid = pyfant.SID(_multi_id_maker())
        self.__runnable_manager = None

    def kill(self):
        self._flag_killed = True
        if self._flag_running and self.__runnable_manager:
            self.__runnable_manager.exit()

    def get_status(self):
        return self.__status

    def run(self):
        """Runs executable.

        Blocking routine. Only returns when executable finishes running.
        """
        assert not self._flag_running, "Already running"
        assert not self._flag_finished, "Already finished"


        self._flag_running = True
        try:
            self.__run()
        except Exception as e:
            self._error_message = e.__class__.__name__+": "+str(e)
            self._flag_error = True
            raise
        finally:
            self._flag_finished = True
            self._flag_running = False
            self.__logger.debug(str(self.__status))

    def __run(self):
        # Called from run() to lower one indentation lever.
        # **Note** If something is not right here: *raise*.
        #
        # **Note** It plays around with Conf, SID, IdMaker objects
        import pyfant as pf

        # # Preparation
        self.__status.stage = "preparing"
        if self.__custom_id:
            self.__sid.id = self.__custom_id
        else:
            self.__sid.make_id()

        # This id maker will create directories inside the
        # multi-session directory.
        # It will replace the Pfant's default id maker
        custom_id_maker = pyfant.IdMaker()
        custom_id_maker.session_prefix_singular = os.path.join(self.__sid.dir, "session-")


        # # Runs innewmarcs and hydro2
        ih = pyfant.Combo([pyfant.FOR_INNEWMARCS, pyfant.FOR_HYDRO2])
        ih.conf.flag_output_to_dir = True
        ih.conf.logger = self.__logger
        ih.conf.opt = copy.copy(self.__options)
        ih.conf.sid = self.__sid
        ih.conf.file_main = copy.copy(self.__file_main)
        # Runs innewmarcs and hydro2;
        # it is also expected to create the "multi-session" directory
        ih.run()

        ####
        # # Runs pfant, pfant creates several .norm files
        self.__status.stage = "pfant stage"
        self.__logger.info("+++ pfant stage...")
        pfant_list = []
        symbols = list(self.__file_abxfwhm.ab.keys())
        abdiffss = list(self.__file_abxfwhm.ab.values())
        n_abdif = len(abdiffss[0])
        fwhms = self.__file_abxfwhm.get_fwhms()
        for fwhm in fwhms:
            if fwhm > 9.99:
                raise RuntimeError("fhwm maximum is 9.99")
        for j in range(n_abdif):
            file_abonds_ = copy.deepcopy(self.__file_abonds)

            for i, symbol in enumerate(symbols):
                found = False
                for k in range(len(file_abonds_.ele)):
                    if file_abonds_.ele[k] == symbol:
                        abdif = abdiffss[i][j]
                        file_abonds_.abol[k] += abdif
                        self.__logger.debug(j, " - ", symbol, "with abundance", file_abonds_.abol[k])
                        found = True

                if not found:
                    raise RuntimeError("Atom '%s' not found" % symbol)

            pfant_name = self.__file_abxfwhm.pfant_names[j] \
                if self.__file_abxfwhm.pfant_names \
                else "%02d" % j
            flprefix = "%s_%s" % (self.__file_main.titrav, pfant_name)

            pfant = pyfant.Pfant()
            pfant.conf.flag_output_to_dir = False
            pfant.conf.opt = copy.copy(self.__options)
            pfant.conf.rename_outputs([pyfant.FOR_INNEWMARCS, pyfant.FOR_HYDRO2], sid=self.sid)
            pfant.conf.sid.id_maker = custom_id_maker
            pfant.conf.sid.id = flprefix
            pfant.conf.opt.flprefix = self.__sid.join_with_session_dir(flprefix)
            pfant.conf.file_main = copy.copy(self.__file_main)
            pfant.conf.file_abonds = file_abonds_
            pfant.conf.file_dissoc = file_abonds_.get_file_dissoc()

            self.__logger.debug("LOOK FLPREFIX "+str(pfant.conf.opt.flprefix))

            pfant_list.append(pfant)
        rm = self.__runnable_manager = pyfant.RunnableManager()
        pyfant.run_parallel(pfant_list, flag_console=False, runnable_manager=rm)
        if self._flag_killed:
            return

        if not rm.flag_success:
            raise RuntimeError("Not all pfant's succeeded running.")

        ####
        # # Runs nulbad, saves .sp and .spl files
        # function to convert given FWHM to string to use as part of a file name
        fmt_fwhm = lambda x: "%03d" % round(x * 100)

        # ## Prepares nulbad's
        nulbad_list = []
        sp_filenames_by_fwhm = {}  # dictionary containing a list of .sp filenames for each FWHM
        for pfant in pfant_list:
            # prefix is sth like: multi-session-1/Sun_00
            prefix = pfant.conf.opt.flprefix

            for fwhm in fwhms:
                nulbad = pyfant.Nulbad()
                nulbad.conf.flag_output_to_dir = False
                nulbad.conf.sid.id_maker = custom_id_maker
                nulbad.conf.opt = copy.copy(self.__options)
                nulbad.conf.opt.fn_flux = pfant.conf.opt.flprefix+".norm"
                nulbad.conf.opt.fwhm = fwhm
                nulbad.conf.opt.fn_cv = "%s_%s.sp" % (prefix, fmt_fwhm(fwhm))
                nulbad_list.append(nulbad)

                if not fwhm in sp_filenames_by_fwhm:
                    sp_filenames_by_fwhm[fwhm] = []
                sp_filenames_by_fwhm[fwhm].append(nulbad.conf.opt.fn_cv)

        # ## Saves files for lineplot.py (lists of spectra)
        # Each item of each list is a full path to a spectrum file
        for fwhm, sp_filenames in list(sp_filenames_by_fwhm.items()):
            spl_filename = os.path.join(self.__sid.dir, "cv_%s.spl" % fmt_fwhm(fwhm))
            with open(spl_filename, "w") as h:
                for sp_filename in sp_filenames:
                    h.write(os.path.abspath(os.path.join(self.__sid.dir, os.path.basename(sp_filename)+"\n")))

        # ## Runs nulbads
        self.__status.stage = "nulbad stage"
        self.__logger.info("+++ nulbad stage...")
        rm = self.__runnable_manager = pyfant.RunnableManager()
        pyfant.run_parallel(nulbad_list, flag_console=False, runnable_manager=rm)
        if self._flag_killed:
            return
        if not rm.flag_success:
            raise RuntimeError("Not all nulbad's succeeded running.")

        ####
        # # Deletes session-* directories if successful
        FLAG_CLEAN = True
        if FLAG_CLEAN:
            self.__logger.info("+++ Cleaning up...")
            for pfant in pfant_list:
                pfant.sid.clean()
            for nulbad in nulbad_list:
                nulbad.sid.clean()
        else:
            self.__logger.info("+++ NOT cleaning up...")

    def _get_sid(self):
        return self.__sid



from mpl_toolkits.mplot3d import Axes3D  # yes, required
import matplotlib.pyplot as plt
import pyfant

__all__ = ["plot_mod_grid"]


def plot_mod_grid(ff, title=None):
    """Plots Teff x glog x metallicity in 3D"""

    fig = plt.figure()
    ax = fig.gca(projection='3d')
    teff, glog, asalog = [], [], []
    for f in ff:
        assert isinstance(f, pyfant.FileModBin)
        for r in f.records:
            teff.append(r.teff)
            glog.append(r.glog)
            asalog.append(r.asalog)
    ax.scatter(teff, glog, asalog, c='r', s=60, marker='o')
    ax.set_xlabel('teff')
    ax.set_ylabel('glog')
    ax.set_zlabel('asalog')
    fig.canvas.set_window_title('teff-glog-asalog scatterplot')
    plt.show()



"""Graphicsl interface for PFANT"""

__all__ = ["XPFANT"]

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import os.path
import copy
import shutil
import a99
import pyfant
from .a_XMainAbonds import *


################################################################################
class XPFANT(XMainAbonds):
    """
    Args:
      parent=None: nevermind
      file_main (optional)-- FileMain instance
    """

    def __init__(self, *args, **kwargs):
        ## State variables
        XMainAbonds.__init__(self, *args, **kwargs)

        self._manager_form = None
        self._rm = None

        # # Central layout

        # ## Main control bar

        l = self.controlLayout
        w = self.buttonSubmit = QPushButton("&Submit job")
        w.clicked.connect(self.on_submit)
        l.addWidget(w)
        w = self.checkbox_custom_id = QCheckBox("Custom session directory")
        w.stateChanged.connect(self.on_checkbox_custom_id_state_changed)
        l.addWidget(w)
        w = self.lineEdit_custom_id = QLineEdit()
        w.setFixedWidth(100)
        l.addWidget(w)
        # s = self.spacer0 = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)
        l.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        # # Final adjustments
        self.setWindowTitle("PFANT launcher")
        self.__update_lineEdit_custom_id()


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # "Duck-typing"

    def set_manager_form(self, x):
        import pyfant as pf
        assert isinstance(x, pyfant.XRunnableManager)
        self._manager_form = x
        self._rm = x.rm

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for Qt library signals

    def on_submit(self):
        flag_ok = True
        errors = self._check_single_setup()
        if len(errors) == 0:
            # more error checking
            if self.checkbox_custom_id.isChecked():
                s = self.__get_custom_session_id()
                if len(s) == 0:
                    errors.append("Please inform custom session id.")
                elif len(errors) == 0: # will only offer to remove directory if everything is ok so far
                    dirname = _get_custom_dirname(s)
                    if os.path.isdir(dirname):
                        r = QMessageBox.question(self, "Directory exists",
                         "Directory '%s' already exists.\n\n"
                         "Would you like to remove it?" % dirname,
                         QMessageBox.Yes|QMessageBox.No, QMessageBox.Yes)
                        if r == QMessageBox.Yes:
                            try:
                                shutil.rmtree(dirname)
                            except Exception as e:
                                errors.append(str(e))
                        else:
                            return

        if len(errors) == 0:
            try:
                self._manager_form.show()
                self._manager_form.raise_()
                self._manager_form.activateWindow()
                self.__submit_job()
            except Exception as e:
                errors.append(str(e))
                a99.get_python_logger().exception("Cannot submit job")
        if len(errors) > 0:
            a99.show_error("Cannot submit job:\n  - "+("\n  - ".join(errors)))

    def on_checkbox_custom_id_state_changed(self):
        self.__update_lineEdit_custom_id()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Slots for signals emited by ftpyfant widgets


    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # Internals

    def __get_custom_session_id(self):
        return str(self.lineEdit_custom_id.text()).strip()

    def __submit_job(self):
        import pyfant as pf
        r = pyfant.Combo()
        if self.checkbox_custom_id.isChecked():
            custom_id = self.__get_custom_session_id()
            r.conf.sid.id = custom_id
            if _get_custom_dirname(custom_id) == custom_id:
                # Understands that session dirname prefix must be cleared
                r.sid.id_maker.session_prefix_singular = ""

        r.conf.opt = copy.copy(self.oe.f)
        r.conf.file_main = self.me.f
        r.conf.file_abonds = self.ae.f
        r.conf.file_dissoc = self.ae.f.get_file_dissoc()
        r.conf.flag_output_to_dir = True
        self._rm.add_runnables([r])

    def __update_lineEdit_custom_id(self):
        self.lineEdit_custom_id.setEnabled(self.checkbox_custom_id.isChecked())


# This sector defines how custom directory name is made up

def _get_custom_dirname(session_id):
    # return pyfant.SESSION_PREFIX_SINGULAR+session_id
    return session_id


"""CO linelist found in Jorge Melendez's 1998 work."""

__all__ = ["FileCOJM1998"]

import a99, math
from f311 import DataFile
from dataclasses import dataclass

@dataclass
class COJM1998Line:
    # wavenumber in cm**-1
    nu: float
    # R2 measured in Debyes**2
    D2: float
    # Einstein's A value in hertz
    A: float

    vl: int
    v2l: int
    # P/Q/R
    branch: str
    J2l: float

    # AFGL isotopologue code
    iso: int


@a99.froze_it
class FileCOJM1998(DataFile):
    """
    Linelist used in Jorge Melendez's 1998 work on CO (filename: "COdv1.dat")
    """

    def __init__(self):
        DataFile.__init__(self)
        self.lines = []

    def __len__(self):
        return len(self.lines)

# Sample from COdv1.dat
# ^^^^^^^^^^^^^^^^^^^^^
#  849.8039 2.928E-01 2.827E+01 69999.5215 1.743E-02 7.914E-33 21 20 P 149 26
#  856.6729 2.918E-01 2.886E+01 69612.2062 1.739E-02 9.572E-33 21 20 P 148 26
#  857.0984 2.876E-01 2.849E+01 68948.3033 1.727E-02 1.265E-32 21 20 P 149 27
#  862.2408 2.837E-01 2.862E+01 68159.3991 1.714E-02 1.796E-32 21 20 P 149 36
# 012345678901234567890123456789012345678901234567890123456789012345678901234
# 0         1         2         3         4         5         6         7
# nu       D2                                                  vl v2 r J2l iso

    def _do_load(self, filename):
        with open(filename, "r") as h:
            for i, s in enumerate(h):
                try:
                    line = COJM1998Line(nu=float(s[0:9]),
                                        D2=float(s[9:19]),
                                        A=float(s[19:29]),
                                        vl=int(s[60:63]),
                                        v2l=int(s[63:66]),
                                        branch=s[67],
                                        J2l=float(s[68:72]),
                                        iso=int(s[72:75]),
                                       )
                    self.lines.append(line)

                except Exception as e:
                    raise RuntimeError("Error around %d%s row of file '%s': \"%s\""%
                                       (i+1, a99.ordinal_suffix(i+1), filename, a99.str_exc(e))) from e



"""TurboSpectrum linelist format. Prepared for molecules only"""


__all__ = ["PlezAtomicLine", "PlezMolecularLine", "PlezSpecies", "FilePlezLinelist", "FilePlezLinelist1",
           "FilePlezLinelistBase", "load_plez_mol", "FilePlezLinelistN14H", "FilePlezLinelist12C16OLi2015",
           "FilePlezLinelistC2"]

import a99
from f311 import DataFile
import io
import os
import re


# Will update progress status every time the following number of lines is read from file
_PROGRESS_INDICATOR_PERIOD = 1030 * 5


def load_plez_mol(filename):
    """
    Attempts to load Plez molecular line list

    Returns:
        FilePlezLinelistBase
    """
    import f311
    return f311.load_with_classes(filename, [FilePlezLinelist, FilePlezLinelist1])


class PlezLine(object):
    def __init__(self):
        self.lambda_ = 1.
        self.chiex = 1.
        self.loggf = 1.
        self.fdamp = 1.
        self.gu = 1.
        self.raddmp = 1.

class PlezAtomicLine(PlezLine):
    pass


class PlezMolecularLine(PlezLine):
    def __init__(self):
        PlezLine.__init__(self)
        self.branch = ""
        self.vl = 0
        self.v2l = 0
        self.Jl = 0.
        self.J2l = 0.


class PlezSpecies(object):
    def __init__(self, elstring, ion, name, lines=None):
        if lines is None:
            lines = []

        self.elstring = elstring
        self.ion = ion
        self.name = name
        self.lines = lines

    def __len__(self):
        return len(self.lines)

    def __iter__(self):
        return iter(self.lines)


class FilePlezLinelistBase(DataFile):
    """Base class for diffent types of Plez atomic/molecular lines file"""


@a99.froze_it
class FilePlezLinelist(FilePlezLinelistBase):
    """
    Plez molecular lines file

    **Note** Plez file so far refers to one molecule only.

    **Note** Load only

    Descentands of this class must implement _do_processline()
    """

    def __init__(self):
        DataFile.__init__(self)

        self.molecules = dict()
        self.atoms = dict()

        self.expr_branch = re.compile("[PQRS]\d*")  # regular expression to find the branch

    def _do_load(self, filename):
        with open(filename, "r") as h:
            self._do_load_h(h, filename)

    def _process_line_molecule(self, species, line, s):
        self._do_process_line_molecule(species, line, s)

    def _do_process_line_molecule(self, species, line, s):
        pass  # 20230605 no need to raise here raise NotImplementedError()

    def _do_load_h(self, h, filename):
        r = 0  # counts rows of file
        ii = 0

        # states
        EXP_SPECIES = 0  # expecting species definition
        EXP_NAME = 1  # expecting name of species
        EXP_ANY = 2

        state = EXP_SPECIES
        is_atom = None
        species = None
        flag_break = False

        try:
            while True:
                s = h.readline().strip()
                if len(s) == 0:
                    flag_break = True

                if not flag_break:
                    if s[0] == "'":
                        if state == EXP_SPECIES or state == EXP_ANY:
                            # New species to come

                            s_ = s[1:]
                            elstring = s_[:s_.index("'")].strip()
                            s__ = s_[s_.index("'")+1:]
                            ion, nline = [int(x.strip()) for x in s__.split(" ") if x.strip()]

                            # Determines whether atom or molecule
                            atnumber = int(s[1:s.index(".")])
                            is_atom = atnumber <= 92

                            if is_atom:
                                elstring = f"{elstring:>02}"
                            else:
                                elstring = f"{elstring:>011}"

                            state = EXP_NAME

                        elif EXP_NAME:
                            name = s.strip("'").strip()
                            species = self._get_species(is_atom, elstring, ion, name)

                            a99.get_python_logger().info(
                                 "Loading '{}': species '{}' ('{}')".format(filename, elstring, name))

                            state = EXP_ANY

                    else:
                        if state != EXP_ANY:
                            raise RuntimeError("Not expecting line right now")
                        # It is a line

                        # Read 6 numeric values then something between quotes
                        line = PlezAtomicLine() if is_atom else PlezMolecularLine()

                        # Description string after 6th value is not mandatory
                        if "'" in s:
                            line.lambda_, line.chiex, line.loggf, line.fdamp, line.gu, line.raddmp = \
                                [float(x) for x in s[:s.index("'")].split()]
                        else:
                            line.lambda_, line.chiex, line.loggf, line.fdamp, line.gu, line.raddmp = \
                                [float(x) for x in s.split()]

                        if not is_atom:
                            self._process_line_molecule(species, line, s)

                        species.lines.append(line)

                if flag_break:
                    break

                r += 1
                ii += 1
                if ii == _PROGRESS_INDICATOR_PERIOD:
                    # a99.get_python_logger().info(
                    #      "Loading '{}': {}".format(filename, a99.format_progress(r, num_lines)))
                    ii = 0

        except Exception as e:
            raise
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename, a99.str_exc(e))) from e

    def _get_species(self, is_atom, elstring, ion, name):
        """Returns PlezSpecies instance, creating one if necessary."""
        list_ = self.atoms if is_atom else self.molecules
        try:
            ret = list_[elstring]
        except KeyError:
            ret = PlezSpecies(elstring, ion, name)
            list_[elstring] = ret
        return ret


    def _do_save_as(self, filename):
        def save_species(h, species):
            a99.write_lf(h, f"'{species.elstring}' {species.ion} {len(species)}")
            a99.write_lf(h, f"'{species.name}'")
            for line in species:
                a99.write_lf(h, f"{line.lambda_:13.3f} {line.chiex:9.3f} {line.loggf:9.3f} {line.fdamp:7.2f} {line.gu:7.1f} {line.raddmp:10.3e}")
        with open(filename, "w") as h:
            for species in self.atoms.values():
                save_species(h, species)
            for species in self.molecules.values():
                save_species(h, species)

    def molecule_by_index(self, i):
        return list(self.molecules.values())[i]

    def atom_by_index(self, i):
        return list(self.atoms.values())[i]

@a99.froze_it
class FilePlezLinelistN14H(FilePlezLinelist):
    """
    Class that can open file "N14H-AX-2011.bsyn.list" (system 'NH A-X PGopher').
    """


    def _do_process_line_molecule(self, species, line, s):

        #'0107.000014          '  1       12150
        #'NH A-X Fernando et al., 2018, JQSRT, 217,29 '
        #   3357.55219    0.004   -2.205 .00    3.  0.000E+00 'qR23(0)        0    1.0  e  F2e  -  0    0.0  e  F3e A            X           '
        #   3349.33946    0.004   -2.625 .00    3.  0.000E+00 'rR3(0)         0    1.0  e  F3e  -  0    0.0  e  F3e       A            X     '

        # Get the branch

        tmp = s[s.index("'"):].strip("'").split()
        _branch = tmp[0]

        line.branch = self.expr_branch.search(_branch).group()
        line.vl = int(tmp[1])
        line.Jl = float(tmp[2])
        line.v2l = int(tmp[6])
        line.J2l = float(tmp[7])


@a99.froze_it
class FilePlezLinelist12C16OLi2015(FilePlezLinelist):
    """
    Class that can open file "12C16O_Hitran.bsyn"

    File "12C16O_Hitran.bsyn" was provided by Masseron in 2023-04 and was converted from Li 2025 HITRAN-format data to
    bsyn (Turbospectrum) data using code so far unknown.

    Complete list of files provided at that time: 12C16O_Hitran.bsyn, 12C17O_Hitran.bsyn, 12C18O_Hitran.bsyn,
    13C16O_Hitran.bsyn, 13C17O_Hitran.bsyn, 13C18O_Hitran.bsyn
    ```
    """

    #'0608.012016          '  1      125497
    #'12C16O Li2015'
    #41558583.29020    8.302   -6.047 .00    3.  0.100E+01 'X' 'X'   0.0    1.0 '     R  0     41   1.0       X1Sigma+  - 41   0.0       X1Sigma+      '
    #40942652.84178    8.161   -6.058 .00    3.  0.100E+01 'X' 'X'   0.0    1.0 '     R  0     40   1.0       X1Sigma+  - 40   0.0       X1Sigma+      '
    #                                                                                 |         |     |                    |     |
    #                                                                            branch        vl    Jl                  v2l   J2l
    #                                                                                 0  1      2     3              4  5  6     7
    #                                                                            0123456789012345678901234567890123456789012345678901234567890123456789
    #                                                                            0         1         2         3         4         5         6

    #'0606.012012 ' 1 24212
    #'EXOMOL '
    #   15000.04111    3.417   -3.130 .00   25.  0.100E+01 'X' 'X'   0.0    1.0 '    23  12.0 F1e  +   b3Sigmag-  - 19  11.0 F2e  -       a3Piu   '
    #   15000.05881    1.724   -8.590 .00   81.  0.100E+01 'X' 'X'   0.0    1.0 '     9  40.0 F1e  +   b3Sigmag-  -  7  39.0 F2e  -       a3Piu   '
    #   15000.15088    1.954   -4.438 .00  139.  0.100E+01 'X' 'X'   0.0    1.0 '     7  69.0 F1f  +   b3Sigmag-  -  6  70.0 F1f  -       a3Piu   '
    #   15000.33324    1.443   -2.043 .00   79.  0.100E+01 'X' 'X'   0.0    1.0 '     5  39.0 F1e  -       A1Piu  -  5  40.0 F0e  +   X1Sigmag+   '
    #   15000.34819    1.954   -4.474 .00  137.  0.100E+01 'X' 'X'   0.0    1.0 '     7  68.0 F0e  +   b3Sigmag-  -  6  69.0 F0e  -       a3Piu   '
    #   15000.36067    2.522   -9.903 .00  185.  0.100E+01 'X' 'X'   0.0    1.0 '     7  92.0 F1e  +   b3Sigmag-  -  2  93.0 F2e  -       a3Piu   '





    # OBS: (2023-04-04) VERY IMPORTANT I think positions 1 and 5 would give further branch information for doublet, triplet etc.


    def _do_process_line_molecule(self, species, line, s):
        # this will give the clean comments section for a line as in sample above
        comments = s.split("'")[5]

        _branch = comments[0:6]
        line.branch = self.expr_branch.search(_branch).group()

        line.vl = int(comments[9:16])
        line.Jl = float(comments[16:22])
        line.v2l = int(comments[40:43])
        line.J2l = float(comments[43:49])


@a99.froze_it
class FilePlezLinelistC2(FilePlezLinelist):
    """
    Deals with .bsyn C2 files (see description).

    This class was made to deal with the following files:
        - 12C12C_15000-17500_P-BR.bsyn
        - 12C13C_15000-17500_P-BR.bsyn
        - 13C13C_15000-17500_P-BR.bsyn
    """

    # '0606.012012 ' 1 24212
    # 'EXOMOL '
    #   15000.04111    3.417   -3.130 .00   25.  0.100E+01 'X' 'X'   0.0    1.0 '    23  12.0 F1e  +   b3Sigmag-  - 19  11.0 F2e  -       a3Piu   '
    #   15000.05881    1.724   -8.590 .00   81.  0.100E+01 'X' 'X'   0.0    1.0 '     9  40.0 F1e  +   b3Sigmag-  -  7  39.0 F2e  -       a3Piu   '
    #   15000.15088    1.954   -4.438 .00  139.  0.100E+01 'X' 'X'   0.0    1.0 '     7  69.0 F1f  +   b3Sigmag-  -  6  70.0 F1f  -       a3Piu   '
    #   15000.33324    1.443   -2.043 .00   79.  0.100E+01 'X' 'X'   0.0    1.0 '     5  39.0 F1e  -       A1Piu  -  5  40.0 F0e  +   X1Sigmag+   '
    #   15000.34819    1.954   -4.474 .00  137.  0.100E+01 'X' 'X'   0.0    1.0 '     7  68.0 F0e  +   b3Sigmag-  -  6  69.0 F0e  -       a3Piu   '
    #   15000.36067    2.522   -9.903 .00  185.  0.100E+01 'X' 'X'   0.0    1.0 '     7  92.0 F1e  +   b3Sigmag-  -  2  93.0 F2e  -       a3Piu   '

    #'0606.012013 ' 1 51113
    #'EXOMOL '
    #   15000.00151    3.598   -5.105 .00  157.  0.100E+01 'X' 'X'   0.0    1.0 '    18  78.0 F0e  +   b3Sigmag-  - 15  79.0 F0e  -       a3Piu   '
    #   15000.00567    0.469   -3.448 .00    3.  0.100E+01 'X' 'X'   0.0    1.0 '     3   1.0 F1f  +   b3Sigmag-  -  2   0.0 F0f  -       a3Piu   '
    #   15000.01264    0.488   -3.181 .00   23.  0.100E+01 'X' 'X'   0.0    1.0 '     3  11.0 F1e  -   b3Sigmag-  -  2  10.0 F0e  +       a3Piu   '
    #   15000.08361    3.396   -2.187 .00   69.  0.100E+01 'X' 'X'   0.0    1.0 '    22  34.0 F0e  +   b3Sigmag-  - 18  34.0 F2f  -       a3Piu   '

    #'0606.013013 ' 1 54388
    #'EXOMOL '
    #   15000.02834    3.306   -2.835 .00   21.  0.100E+01 'X' 'X'   0.0    1.0 '    23  10.0 F1f  -   b3Sigmag-  - 19  10.0 F0e  +       a3Piu   '
    #   15000.06192    1.757   -6.268 .00   95.  0.100E+01 'X' 'X'   0.0    1.0 '     9  47.0 F1f  +   b3Sigmag-  -  7  47.0 F0e  -       a3Piu   '
    #   15000.09163    1.087   -6.495 .00  151.  0.100E+01 'X' 'X'   0.0    1.0 '     1  75.0 F0e  -   b3Sigmag-  -  0  75.0 F1f  +       a3Piu   '
    #   15000.15101    2.122   -6.552 .00  135.  0.100E+01 'X' 'X'   0.0    1.0 '     9  67.0 F0e  -   b3Sigmag-  -  8  66.0 F0e  +       a3Piu   '
    #                                                                                 |     |                        |     |
    #                                                                                vl    Jl                      v2l   J2l
    #                                                                            0123456789012345678901234567890123456789012345678901234567890123456789
    #                                                                            0         1         2         3         4         5         6

    def _do_process_line_molecule(self, species, line, s):
        # this will give the clean comments section for a line as in sample above
        comments = s.split("'")[5]

        line.branch = "-"

        row = [x for x in comments.split(" ") if x]
        line.vl = int(row[0])
        line.Jl = float(row[1])
        line.v2l = int(row[6])
        line.J2l = float(row[7])
        line.from_label = row[4][0]
        line.to_label = row[10][0]
        # line.vl = int(comments[0:6])
        # line.Jl = float(comments[6:12])
        # line.v2l = int(comments[34:37])
        # line.J2l = float(comments[37:43])
        # line.from_label = comments[22]
        # line.to_label = comments[57]


@a99.froze_it
class FilePlezLinelist1(FilePlezLinelistBase):
    """
    Plez molecular lines file having "linelistCN1214V130710.dat" as holotype.
    """

    def __init__(self):
        DataFile.__init__(self)

        self.molecules = {}
        self.atoms = {}

    def _do_load(self, filename):
        with open(filename, "r") as h:
            self._do_load_h(h, filename)

    def _do_load_h(self, h, filename):
        r = 0  # counts rows of file
        ii = 0
        species = None
        flag_break = False

        try:
            while True:
                s = h.readline().strip()
                if len(s) == 0:
                    break

                # linelistCN1214V130710.dat
                #
                # Mol vup vlow JLow branch gf       wavenumber  exc (eV) transition (V=violet, R=red)
                # 0   1   2    3    5      6        7
                #---BEGIN SAMPLE---
                #CN1 11 03 034 +  R1   0.4853E-07  40369.070  1.020 V
                #CN1 11 03 035 +  R1   0.5286E-07  40359.670  1.036 V
                #CN1 11 03 036 +  R1   0.5757E-07  40349.879  1.052 V
                #---END SAMPLE---

                tmp = s.split()
                species = self.molecules[tmp[0]]

                line = PlezMolecularLine()

                line.Jl = None
                line.vl = int(tmp[1])
                line.v2l = int(tmp[2])
                line.J2l = float(tmp[3])
                line.lambda_ = 1/float(tmp[7])*1e8  # we assume the wavenumber is in 1/cm
                line.branch = tmp[5]

                species.lines.append(line)

                r += 1
                ii += 1
                if ii == _PROGRESS_INDICATOR_PERIOD:
                    # a99.get_python_logger().info(
                    #      "Loading '{}': {}".format(filename, a99.format_progress(r, num_lines)))
                    ii = 0

        except Exception as e:
            raise RuntimeError("Error around %d%s row of file '%s': \"%s\"" %
                               (r + 1, a99.ordinal_suffix(r + 1), filename, a99.str_exc(e))) from e




from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import a99
import copy
import pyfant
import tabulate

__all__ = ["WMolecularConstants"]


# Max J2l / "JJ" for the HLF window
MAX_JJ = 200

# Spacing for grid layouts
_LAYSP_GRID = 4
# Margin for grid layouts
_LAYMN_GRID = 0
_SPACE_BETWEEN_FIELDS = 8
# Spacing between label and combobox
_LAYSP_CB = 3
# Spacing for vertical layouts
_LAYSP_V = 10
# Margin for vertical layouts ((caption, combobox); grid)
_LAYMN_V = 6
# Spacing for main layout
_LAYSP_MAIN = 6
# Margin for main layout
_LAYMN_MAIN = 0
# Spacing for toolbar
_LAYSP_TOOLBAR = 8
# Margin for toolbar
_LAYMN_TOOLBAR = 2
# Spacing for frames that are inside frames
_LAYSP_FRAME1 = 6
# Margin for frames inside frames
_LAYMN_FRAME1 = 0



################################################################################
class _XHLF(a99.XLogMainWindow):
    """
    Window to show Hönl-London factors for given MolConsts object
    Args:
        parent: nevermind
        moldb:
        molconsts:
        fcfs

    """

    def __init__(self, parent, moldb, molconsts, fcfs):
        a99.XLogMainWindow.__init__(self, parent)

        self.moldb = moldb
        self.molconsts = molconsts
        self.fcfs = fcfs

        cw = self.centralWidget = QWidget()
        self.setCentralWidget(cw)


        # # Vertical layout: toolbar and text are stacked

        lv = self.layout_main = QVBoxLayout(cw)

        # ## Horizontal layout: toolbar

        lh = self.layout_molecule = QHBoxLayout()
        lv.addLayout(lh)

        ###
        la = self.keep_ref(QLabel("Multiply by Franck-Condon Factors"))
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        lh.addWidget(la)
        ###
        cb = self.checkbox_fcf = QCheckBox()
        lh.addWidget(cb)

        ###
        b = self.keep_ref(QPushButton("&Calculate"))
        lh.addWidget(b)
        b.clicked.connect(self._calculate)
        ###
        lh.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        ###

        # ## Text area

        cw = self.textEdit = QPlainTextEdit()
        lv.addWidget(cw)
        cw.setReadOnly(True)  # allows copy but not editing
        cw.setFont(a99.MONO_FONT)

        self.setWindowTitle("Hönl-London factors")
        self.setGeometry(0, 0, 800, 600)
        a99.place_center(self)
        a99.nerdify(self)

    def _calculate(self):

        flag_fcf = self.checkbox_fcf.isChecked()

        if flag_fcf:
            if self.fcfs is None or len(self.fcfs) == 0:
                self.add_log_error("FCFs not available", True)

        try:

            l_text = []

            self.status("Calculating...")
            molconsts = copy.deepcopy(self.molconsts)
            molconsts.None_to_zero()
            for vl in range(10):
                for v2l in range(10):
                    l_text.extend(a99.format_box("vl, v2l = ({}, {})".format(vl, v2l)))

                    factor = 1.
                    if flag_fcf:
                        try:
                            factor = self.fcfs[(vl, v2l)]
                        except KeyError:
                            l_text.append("Franck-Condon Factor not available")
                            continue

                    rows, header = [], None
                    for J in range(MAX_JJ+1):
                        mtools = pyfant.kovacs_toolbox(molconsts, flag_normalize=True)
                        mtools.populate(vl, v2l, J)

                        if header is None:
                            branches = [key[3] for key in mtools.dict_sj]
                            header = ["J"]+branches+["Sum"]

                        vv = mtools.dict_sj.values()
                        total = sum([x*factor for x in vv if x != pyfant.NO_LINE_STRENGTH])
                        rows.append([J+.5]+["{:.5e}".format(x*factor) if x != pyfant.NO_LINE_STRENGTH else "-" for x in vv]+["{:.9g}".format(total)])

                    l_text.append(tabulate.tabulate(rows, header))
                    l_text.append("")

            text = "\n".join(l_text)
            self.textEdit.setPlainText(text)
            self.status("")
        except Exception as E:
            self.add_log_error(a99.str_exc(E), True)
            a99.get_python_logger().exception("Error calculating HLF")

"""
        sbx = self.spinBox_redshift = QDoubleSpinBox()
        laa.setBuddy(sbx)
        sbx.setSingleStep(.01)
        sbx.setDecimals(2)
        sbx.setMinimum(-.5)
        sbx.setMaximum(17)
        # sbx.valueChanged.connect(self.spinBoxValueChanged)
        lwset.addWidget(sbx)

"""

class _XTRAPRBInput(a99.XLogMainWindow):
    """
    Window to show Hönl-London factors for given MolConsts object
    Args:
      parent: nevermind
      molconsts:
    """

    def __init__(self, parent, molconsts):
        a99.XLogMainWindow.__init__(self, parent)

        self.molconsts = molconsts

        cw = self.centralWidget = QWidget()
        self.setCentralWidget(cw)


        # # Vertical layout: toolbar and text are stacked

        lv = self.layout_main = QVBoxLayout(cw)

        # ## Horizontal layout: toolbar

        lh = self.layout_toolbar = QHBoxLayout()
        lv.addLayout(lh)


        ###
        TT = "Maximum vibrational quantum number for calculation of Franck-Condon factors"
        la = self.keep_ref(QLabel("Maximum v'' (MAXV)"))
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        lh.addWidget(la)
        ###
        sbx = self.spinBox_vmax = QSpinBox()
        la.setBuddy(sbx)
        sbx.setSingleStep(1)
        sbx.setMinimum(2)
        sbx.setMaximum(30)
        sbx.setValue(12)  # default maxv in TRAPRBInputState.__init__()
        lh.addWidget(sbx)

        ###
        b = self.keep_ref(QPushButton("&Generate"))
        lh.addWidget(b)
        b.clicked.connect(self._calculate)
        ###
        lh.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
        ###

        # ## Text area

        cw = self.textEdit = QPlainTextEdit()
        lv.addWidget(cw)
        cw.setReadOnly(True)  # allows copy but not editing
        cw.setFont(a99.MONO_FONT)

        self.setWindowTitle(a99.get_obj_doc0(pyfant.FileTRAPRBInput))
        self.setGeometry(0, 0, 800, 600)
        a99.place_center(self)
        a99.nerdify(self)

    def _calculate(self):
        f = pyfant.FileTRAPRBInput()
        try:
            f.from_molconsts(self.molconsts, maxv=self.spinBox_vmax.value())
        except Exception as e:
            self.add_log_error("Could not generate TRAPRB input: {}".format(a99.str_exc(e)), True, e)
        else:
            text = f.dumps()
            self.textEdit.setPlainText(text)


class WMolecularConstants(a99.WBase):
    """
    Widget for the user to type or obtain many molecular constants
    """

    @property
    def fn_moldb(self):
        return self._get_fn_moldb()

    @fn_moldb.setter
    def fn_moldb(self, x):
        self._fn_moldb = x
        self._update_gui()

    @property
    def moldb(self):
        return self._moldb

    @property
    def molconsts(self):
        return self._molconsts

    @molconsts.setter
    def molconsts(self, molconsts):
        self._molconsts = molconsts
        self._update_gui()

    @property
    def fcfs(self):
        """Returns a dictionary of Franck-Condon Factors (key is (vl, v2l)), or None"""
        return self._get_fcf_dict()

    @property
    def id_molecule(self):
        return self._get_id_molecule()

    @property
    def flag_valid(self):
        return self._flag_valid

    # @property
    # def row(self):
    #     """Wraps WDBMolecule.row"""
    #     return self.w_mol.row

    def __init__(self, *args):
        a99.WBase.__init__(self, *args)

        self._molconsts = pyfant.MolConsts()
        self._flag_valid = True

        # activated when populating table
        self._flag_populating = False
        # # _flag_populating_* collection
        self._flag_populating_molecule = False
        self._flag_populating_system = False
        self._flag_populating_states = False
        self._flag_populating_pfantmol = False
        self._flag_updating_gui = False

        # activated when searching for statel, state2l
        self._flag_searching_states = False
        # activated when searching for pfantmol
        self._flag_searching_pfantmol = False
        # activated when searching for system
        self._flag_searching_system = False

        # # Internal state

        # FileMolDB object, auxiliary object needed for operations
        self._moldb = None

        # Fields of interest from table 'pfantmol'
        # TODO cro disabled but still left to get rid of this field in the same way I got rid of pfantmol.s
        self._fieldnames_pfantmol = ["fe", "do", "am", "bm", "ua", "ub", "te", ]
        # Fields of interest from table 'state'
        self._fieldnames_state = ["omega_e", "omega_ex_e", "omega_ey_e", "B_e", "alpha_e", "D_e",
                                 "beta_e", "A"]
        # Fields of interest from table 'system'
        self._fieldnames_system = ["from_label", "from_mult", "from_spdf", "to_label", "to_mult", "to_spdf"]
        self._flag_built_edits = False
        for fn in self._fieldnames_state:
            assert fn not in self._fieldnames_pfantmol
            assert fn not in self._fieldnames_system
        for fn in self._fieldnames_pfantmol:
            assert fn not in self._fieldnames_system
        self._fieldnames = []  # will be filled in later copy.copy(self._fieldnames_pfantmol)
        # self._fieldnames.extend(self._fieldnames_state)
        # self._fieldnames.extend(self._fieldnames_system)

        # dictionary {(field name): (edit object), }
        # (will be populated later below together with edit widgets creation)
        self._edit_map = {}
        self._edit_map_statel = {}
        self._edit_map_state2l = {}
        self._edit_map_pfantmol = {}
        self._edit_map_system = {}

        # id from table 'molecule'. In sync with combobox_molecule items
        self._ids_molecule = []
        # id from table 'pfantmol'. In sync with combobox_pfantmol
        self._ids_pfantmol = []
        # id from table 'state'. In sync with combobox_statel and combobox_state2l
        self._ids_state = []
        # id from table 'system'. In sync with combobox_system
        self._ids_system = []

        # # GUI design

        l = self.layout_main = QVBoxLayout(self)
        a99.set_margin(l, _LAYMN_MAIN)
        l.setSpacing(_LAYSP_MAIN)


        # ## Select molecule combobox
        l0 = self.layout_molecule = QHBoxLayout()
        l.addLayout(l0)
        l0.setSpacing(_LAYSP_CB)
        la = self.label_molecule = QLabel()
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l0.addWidget(la)
        cb = self.combobox_molecule = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_molecule_currentIndexChanged)
        l0.addWidget(cb)

        la.setBuddy(cb)

        # ## Frame with the combobox to select the "system"

        fr = self.frame_system = a99.get_frame()
        l.addWidget(fr)
        l1 = self.layout_frame_system = QVBoxLayout(fr)
        l1.setSpacing(_LAYSP_V)
        a99.set_margin(l1, _LAYMN_V)

        # ### Select system combobox
        l2 = self.layout_system = QHBoxLayout()
        l1.addLayout(l2)
        l2.setSpacing(_LAYSP_CB)
        la = self.label_system = QLabel()
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l2.addWidget(la)
        cb = self.combobox_system = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_system_currentIndexChanged)
        l2.addWidget(cb)

        la.setBuddy(cb)

        la = self.label_fcf = QLabel()
        la.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed))
        l2.addWidget(la)


        # ### State constants edit fields

        lg = self.layout_grid_system = QGridLayout()
        l1.addLayout(lg)
        a99.set_margin(lg, _LAYMN_GRID)
        lg.setSpacing(_LAYSP_GRID)

        # #### Frame for pfantmol combobox end edit fields


        lfp = self.layout_outer_frame_pfantmol = QVBoxLayout()
        l1.addLayout(lfp)
        lfp.setSpacing(_LAYSP_FRAME1)
        lfp.setContentsMargins(_LAYMN_FRAME1, 0, _LAYMN_FRAME1, _LAYMN_FRAME1, )

        fr = self.frame_pfantmol = a99.get_frame()
        lfp.addWidget(fr)
        l5 = self.layout_frame_pfantmol = QVBoxLayout(fr)
        l5.setSpacing(_LAYSP_V)
        a99.set_margin(l5, _LAYMN_V)

        # ##### Label and combobox in H layout
        l55 = self.layout_pfantmol = QHBoxLayout()
        l5.addLayout(l55)
        l55.setSpacing(_LAYSP_CB)

        la = self.label_pfantmol = QLabel("<b>&PFANT molecule</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l55.addWidget(la)

        cb = self.combobox_pfantmol = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_pfantmol_currentIndexChanged)
        l55.addWidget(cb)

        la.setBuddy(cb)

        # ##### PFANT Molecular constants edit fields
        lg = self.layout_grid_pfantmol = QGridLayout()
        a99.set_margin(lg, _LAYMN_GRID)
        lg.setSpacing(_LAYSP_GRID)
        l5.addLayout(lg)


        # #### Statel and State2l

        fr = self.frame_states = a99.get_frame()
        self.layout_main.addWidget(fr)

        l33 = self.layout_frame_states = QVBoxLayout(fr)
        l33.setSpacing(_LAYSP_V)
        a99.set_margin(l33, _LAYMN_V)

        la = self.label_states = QLabel("<b>Constants for individual states</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l33.addWidget(la)

        # This layout will have two frames layed horizontally
        lll = self.layout_states = QHBoxLayout()
        lll.setSpacing(_LAYSP_FRAME1)
        lll.setContentsMargins(_LAYMN_FRAME1, 0, _LAYMN_FRAME1, _LAYMN_FRAME1, )
        l33.addLayout(lll)

        # ##### Frame for Diatomic molecular constants for statel

        fr = self.frame_statel = a99.get_frame()
        lll.addWidget(fr)
        l3 = self.layout_frame_statel = QVBoxLayout(fr)
        l3.setSpacing(_LAYSP_V)
        a99.set_margin(l3, _LAYMN_V)

        # ###### Select statel combobox
        l4 = self.layout_statel = QHBoxLayout()
        l3.addLayout(l4)
        l4.setSpacing(_LAYSP_CB)
        la = self.label_statel = QLabel("<b>Diatomic molecular constants for state'</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l4.addWidget(la)
        cb = self.combobox_statel = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_statel_currentIndexChanged)
        l4.addWidget(cb)

        la.setBuddy(cb)

        # ###### Statel edit fields

        lg = self.layout_grid_statel = QGridLayout()
        a99.set_margin(lg, _LAYMN_GRID)
        lg.setSpacing(_LAYSP_GRID)
        l3.addLayout(lg)


        # ##### Frame for Diatomic molecular constants for state2l

        fr = self.frame_state2l = a99.get_frame()
        lll.addWidget(fr)
        l3 = self.layout_frame_state2l = QVBoxLayout(fr)
        l3.setSpacing(_LAYSP_V)
        a99.set_margin(l3, _LAYMN_V)

        # ###### Select state2l combobox
        l4 = self.layout_state2l = QHBoxLayout()
        l3.addLayout(l4)
        l4.setSpacing(_LAYSP_CB)
        la = self.label_state2l = QLabel("<b>Diatomic molecular constants for state''</b>")
        la.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed))
        l4.addWidget(la)
        cb = self.combobox_state2l = QComboBox()
        cb.currentIndexChanged.connect(self.combobox_state2l_currentIndexChanged)
        l4.addWidget(cb)

        la.setBuddy(cb)

        # ###### State2l edit fields

        lg = self.layout_grid_state2l = QGridLayout()
        a99.set_margin(lg, _LAYMN_GRID)
        lg.setSpacing(_LAYSP_GRID)
        l3.addLayout(lg)


        # ## Toolbar at bottom
        self.layout_toolbar = self._get_toolbar_bottom()

        a99.nerdify(self)

    def _get_toolbar_bottom(self):
        ltb = QHBoxLayout()
        self.layout_main.addLayout(ltb)
        a99.set_margin(ltb, _LAYMN_TOOLBAR)
        ltb.setSpacing(_LAYSP_TOOLBAR)
        bb = self.button_zeros = QPushButton("Fill empty fields with zeros")
        bb.clicked.connect(self.None_to_zero)
        ltb.addWidget(bb)

        ltb.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))

        bb = self.button_zeros = QPushButton("View Hönl-London factors")
        bb.clicked.connect(self.view_hlf)
        ltb.addWidget(bb)

        bb = self.button_zeros = QPushButton("Generate TRAPRB input")
        bb.setToolTip("Generates text to be used as input file for the TRAPRB Fortran code [Jarmain&McCallum1970]\n"
                      "This code calculates Franck-Condon factors")
        bb.clicked.connect(self.view_traprb)
        ltb.addWidget(bb)

        return ltb

    def __getitem__(self, fieldname):
        return self.molconsts[fieldname]

    def __setitem__(self, fieldname, value):
        """Can assign widget properties by key. TODO Does not work with id_* yet"""
        if fieldname not in self._fieldnames:
            raise KeyError("Invalid fieldname '{}'".format(fieldname))

        widget = self._edit_map[fieldname]
        text = str(value) if value is not None else ""
        widget.setText(text)


    def set_moldb(self, fobj):
        self._set_moldb(fobj)

    def None_to_zero(self):
        """Fills missing values with zeros"""
        for edit in self._edit_map.values():
            if len(edit.text().strip()) == 0:
                edit.setText("0")

        self._update_molconsts()

    def view_hlf(self):
        """Opens text window with HLFs on them"""


        if self.moldb is None:
            a99.show_error("Cannot view HLFs because I need a molecular constants database")
        elif self.molconsts is None:
            a99.show_error("Cannot view HLFs because I am blank")
        else:
            form = _XHLF(self, self.moldb, self.molconsts, self._get_fcf_dict())
            form.show()

    def view_traprb(self):
        """Opens text window TRAPRB input as text"""

        if not self._can_calculate():
            return

        form = _XTRAPRBInput(self, self.molconsts)
        form.show()

    def _can_calculate(self):
        """Shows error if the case. Returns True/False"""
        reason = None
        if self.moldb is None:
            reason = "I need a molecular constants database"
        elif self.molconsts is None:
            reason = "I am blank"
        if reason:
            a99.show_error(reason)
        return reason is None

    ################################################################################################
    # # Qt override

    def eventFilter(self, obj_focused, event):
        if event.type() == QEvent.FocusIn:
            if obj_focused in self._edit_map.values():
                self.status(obj_focused.toolTip())
        return False

    ################################################################################################
    # # Slots

    def _onTextEdited(self):
        self._update_molconsts()
        self.changed.emit()

    def combobox_molecule_currentIndexChanged(self):
        if self._flag_updating_gui:
            return
        if self._flag_populating_molecule:
            return

        self._populate_sub_comboboxes()
        self._auto_search()
        self._update_molconsts()
        self.changed.emit()

    def combobox_system_currentIndexChanged(self):
        if self._flag_updating_gui:
            return
        if self._flag_populating_system:
            return
        if self._flag_searching_system:
            return

        self._fill_edits_system()
        self._update_label_fcf()

        self._populate_combobox_pfantmol()

        self._auto_search_states()
        self._auto_search_pfantmol()
        self._update_molconsts()
        self.changed.emit()

    def combobox_pfantmol_currentIndexChanged(self):
        if self._flag_updating_gui:
            return
        if self._flag_populating_pfantmol:
            return
        if self._flag_searching_pfantmol:
            return

        self._fill_edits_pfantmol()
        self._update_molconsts()
        self.changed.emit()

    def combobox_statel_currentIndexChanged(self):
        if self._flag_updating_gui:
            return
        if self._flag_populating_states:
            return
        if self._flag_searching_states:
            return

        self._fill_edits_statel()
        self._update_molconsts()
        self.changed.emit()

    def combobox_state2l_currentIndexChanged(self):
        if self._flag_updating_gui:
            return
        if self._flag_populating_states:
            return
        if self._flag_searching_states:
            return

        self._fill_edits_state2l()
        self._update_molconsts()
        self.changed.emit()

    ################################################################################################
    # # Internal function

    def _get_id_molecule(self):
        idx = self.combobox_molecule.currentIndex()
        if idx > 0:
            return self._ids_molecule[idx-1]
        return None

    def _get_id_pfantmol(self):
        idx = self.combobox_pfantmol.currentIndex()
        if idx > 0:
            return self._ids_pfantmol[idx-1]
        return None

    def _get_id_statel(self):
        idx = self.combobox_statel.currentIndex()
        if idx > 0:
            return self._ids_state[idx-1]
        return None

    def _get_id_state2l(self):
        idx = self.combobox_state2l.currentIndex()
        if idx > 0:
            return self._ids_state[idx-1]
        return None

    def _get_id_system(self):
        idx = self.combobox_system.currentIndex()
        if idx > 0:
            return self._ids_system[idx - 1]
        return None

    def _get_fcf_dict(self):
        if self.moldb is None:
            return None
        return self.moldb.get_fcf_dict(self._get_id_system())

    def _set_moldb(self, fobj):
        # assert isinstance(fobj, pyfant.FileMolDB), "I dont want a {}".format(fobj)
        self._moldb = fobj
        self._populate()
        self._update_gui()
        # self._auto_search()

    def _populate(self):
        if not self._flag_built_edits and self.moldb is not None:
            self._build_edits()
        self._populate_combobox_molecule()
        self._populate_sub_comboboxes()

    def _populate_sub_comboboxes(self):
        self._populate_combobox_system()
        self._populate_combobox_pfantmol()
        self._populate_combobox_state()

    def _populate_combobox_molecule(self):
        if self._flag_populating_molecule:
            return

        self._flag_populating_molecule = True
        try:
            cb = self.combobox_molecule
            cb.clear()
            self._ids_molecule = []
            data = self.moldb.query_molecule().fetchall() if self.moldb is not None else []

            cb.addItem("(please select)" if len(data) > 0 else "(no data)")

            for row in data:
                    cb.addItem("{:10} {}".format(row["formula"], row["name"]))
                    self._ids_molecule.append(row["id"])
            self._set_caption_molecule()

        finally:
            self._flag_populating_molecule = False

    def _populate_combobox_pfantmol(self):
        if self._flag_populating_pfantmol:
            return

        self._flag_populating_pfantmol = True
        try:
            cb = self.combobox_pfantmol
            cb.clear()
            self._ids_pfantmol = []

            id_system = self._get_id_system()

            data = [] if id_system is None else \
                self.moldb.query_pfantmol(id_system=self._get_id_system()).fetchall()

            cb.addItem("(please select)" if len(data) > 0 else "(no data)")

            for row in data:
                cb.addItem("{}".format(row["description"]))
                self._ids_pfantmol.append(row["id"])

            self._fill_edits_pfantmol()
            self._set_caption_pfantmol()
        finally:
            self._flag_populating_pfantmol = False

    def _populate_combobox_state(self):
        if self._flag_populating_states:
            return

        self._flag_populating_states = True
        try:
            self._ids_state = []
            cb0 = self.combobox_statel
            cb0.clear()
            cb1 = self.combobox_state2l
            cb1.clear()
            data = [] if self.moldb is None else \
                self.moldb.query_state(id_molecule=self._get_id_molecule()).fetchall()

            cb0.addItem("(please select)" if len(data) > 0 else "(no data)")
            cb1.addItem("(please select)" if len(data) > 0 else "(no data)")

            for row in data:
                cb0.addItem("{}".format(row["State"]))
                self._ids_state.append(row["id"])

            for row in data:
                cb1.addItem("{}".format(row["State"]))

            self._set_caption_state()

        finally:
            self._flag_populating_states = False

    def _populate_combobox_system(self):
        if self._flag_populating_system:
            return

        self._flag_populating_system = True
        try:
            cb = self.combobox_system
            cb.clear()
            self._ids_system = []
            data = [] if self.moldb is None else \
                self.moldb.query_system(id_molecule=self._get_id_molecule()).fetchall()
            cb.addItem("(please select)" if len(data) > 0 else "(no data)")
            for row in data:
                cb.addItem(pyfant.molconsts_to_system_str(row, style=pyfant.SS_ALL_SPECIAL))
                self._ids_system.append(row["id"])

            self._fill_edits_system()
            self._update_label_fcf()
            self._set_caption_system()

        finally:
            self._flag_populating_system = False

    def _build_edits(self):
        self._build_edits_generic(6, self._fieldnames_system, "system", self.layout_grid_system, self._edit_map_system)
        self._build_edits_generic(4, self._fieldnames_pfantmol, "pfantmol", self.layout_grid_pfantmol, self._edit_map_pfantmol)
        self._build_edits_generic(4, self._fieldnames_state, "state", self.layout_grid_statel, self._edit_map_statel, "statel_")
        self._build_edits_generic(4, self._fieldnames_state, "state", self.layout_grid_state2l, self._edit_map_state2l, "state2l_")
        self._flag_built_edits = True

    def _build_edits_generic(self, nc, fieldnames, tablename, layout, edit_map, fieldname_all_prefix=""):

        def new_spacer():
            w = QWidget(self)
            self.keep_ref(w)
            #w.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum))
            w.setFixedWidth(_SPACE_BETWEEN_FIELDS)
            return w



        n = len(fieldnames)
        ti = self.moldb.get_table_info(tablename)
        for k in range(n):
            i = k // nc
            j = k % nc

            fieldname = fieldnames[k]
            fieldname_all = fieldname_all_prefix+fieldname

            info = ti[fieldname]
            caption = info["caption"] or fieldname
            a = QLabel(caption)
            e = QLineEdit("")
            e.textEdited.connect(self._onTextEdited)
            e.installEventFilter(self)
            tooltip = info["tooltip"]
            if tooltip:
                a.setToolTip(tooltip)
                e.setToolTip(tooltip)
            self._fieldnames.append(fieldname_all)
            self._edit_map[fieldname_all] = e
            edit_map[fieldname] = e
            layout.addWidget(a, i, j * 3)
            layout.addWidget(e, i, j * 3 + 1)
            if j < nc-1:
                layout.addWidget(new_spacer(), i, j*3+2)

    def _fill_edits_pfantmol(self):
        id_ = self._get_id_pfantmol()
        if id_ is not None:
            row = self.moldb.query_pfantmol(**{"pfantmol.id": id_}).fetchone()
            for fieldname, e in self._edit_map_pfantmol.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")

    def _fill_edits_statel(self):
        id_ = self._get_id_statel()
        if id_ is not None:
            row = self.moldb.query_state(**{"state.id": id_}).fetchone()
            for fieldname, e in self._edit_map_statel.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")

    def _fill_edits_state2l(self):
        id_ = self._get_id_state2l()
        if id_ is not None:
            row = self.moldb.query_state(**{"state.id": id_}).fetchone()
            for fieldname, e in self._edit_map_state2l.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")


    def _fill_edits_system(self):
        id_ = self._get_id_system()
        if id_ is not None:
            row = self.moldb.query_system(**{"system.id": id_}).fetchone()
            for fieldname, e in self._edit_map_system.items():
                v = row[fieldname]
                e.setText(str(v) if v is not None else "")

    def _update_label_fcf(self):
        n, fcfs = 0, self._get_fcf_dict()
        if fcfs is not None:
            n = len(fcfs)
        # Left space to take a distance from combobox left to it
        s = '<font color="{}">{}Franck-Condon Factors (FCFs) for {} vibrational transition{}</font>'.\
            format("blue" if n != 0 else "red",
                   "",
                   n,
                   "" if n == 1 else "s")
        self.label_fcf.setText(s)

    def _auto_search(self):
        self._auto_search_system()
        self._populate_combobox_pfantmol()
        self._auto_search_pfantmol()
        self._auto_search_states()

    def _auto_search_system(self):
        if self._flag_populating_system or self._flag_searching_system:
            return

        self._flag_searching_system = True
        try:
            if len(self._ids_system) > 0:
                self.combobox_system.setCurrentIndex(1)
                self._fill_edits_system()
                self._update_label_fcf()
        finally:
            self._flag_searching_system = False

    def _auto_search_pfantmol(self):
        if self._flag_populating_pfantmol or self._flag_searching_pfantmol:
            return

        self._flag_searching_pfantmol = True
        try:
            if len(self._ids_pfantmol) > 0:
                self.combobox_pfantmol.setCurrentIndex(1)
                self._fill_edits_pfantmol()
        finally:
            self._flag_searching_pfantmol = False

    def _auto_search_states(self):

        if self._flag_populating_states or self._flag_searching_states:
            return

        self._flag_searching_states = True
        try:
            self.__auto_search_state("from_label")
            self.__auto_search_state("to_label")
        finally:
            self._flag_searching_states = False

    def __auto_search_state(self, fieldname="from_label"):
        cb = self.combobox_statel if fieldname == "from_label" else self.combobox_state2l
        id_molecule = self._get_id_molecule()
        if id_molecule is not None:
            id_system = self._get_id_system()
            if id_system is not None:
                row_system = self.moldb.query_system(id=id_system).fetchone()

                state = row_system[fieldname]
                row_state = self.moldb.get_conn().execute("select * from state where id_molecule = ? "
                 "and substr(State, 1, {}) = ?".format(len(state)), (id_molecule, state,)).fetchone()

                if row_state is not None:
                    cb.setCurrentIndex(self._ids_state.index(row_state["id"]) + 1)
                    if fieldname == "from_label":
                        self._fill_edits_statel()
                    else:
                        self._fill_edits_state2l()

    def _set_caption_molecule(self):
        self.label_molecule.setText("<b>&Molecule ({})</b>".format(len(self._ids_molecule)))


    def _set_caption_system(self):
        self.label_system.setText("<b>&Electronic system ({})</b>".format(len(self._ids_system)))


    def _set_caption_state(self):
        self.label_statel.setText("<b>State ' ({})</b>".format(len(self._ids_state)))
        self.label_state2l.setText("<b>State '' ({})</b>".format(len(self._ids_state)))

    def _set_caption_pfantmol(self):
        self.label_pfantmol.setText("<b>&PFANT molecule ({})</b>".format(len(self._ids_pfantmol)))


    def _update_gui(self):
        if self._flag_updating_gui:
            return

        def populate_two_comboboxes():
            self._populate_combobox_system()
            self._populate_combobox_state()


        self._flag_updating_gui = True

        try:
            molconsts = self._molconsts

            # map to update the comboboxes
            _CB_MAP = [
                ("id_molecule", self.combobox_molecule, lambda: self._ids_molecule,
                 populate_two_comboboxes),
                ("id_system", self.combobox_system, lambda: self._ids_system,
                 self._populate_combobox_pfantmol),
                ("id_pfantmol", self.combobox_pfantmol, lambda: self._ids_pfantmol, None),
                ("id_statel", self.combobox_statel, lambda: self._ids_state, None),
                ("id_state2l", self.combobox_state2l, lambda: self._ids_state, None),
            ]

            # # Stage 1/2: comboboxes
            for fieldname, cb, f_ids, method_after in _CB_MAP:
                value = molconsts[fieldname]
                if value is not None:
                    try:
                        idx = f_ids().index(value)
                        cb.setCurrentIndex(idx+1)

                        if method_after is not None:
                            method_after()

                    except ValueError:
                        pass  # nevermind, just won't position combobox properly


            # # Stage 2/2: edits
            for fieldname in self._fieldnames:
                self[fieldname] = molconsts[fieldname]

        finally:
            self._update_label_fcf()
            self._flag_updating_gui = False

    def _update_molconsts(self):
        """Updates molconsts from GUI"""

        emsg, flag_error = "", False
        fieldname = None
        try:
            # edits
            for fieldname in self._fieldnames:
                text = self._edit_map[fieldname].text()
                if "_label" in fieldname:
                    # Fields with "_label" in their names are treated as strings,
                    # otherwise they are considered numeric
                    value = text
                else:
                    if len(text.strip()) > 0:
                        value = float(text)
                    else:
                        value = None

                self.molconsts[fieldname] = value

            # comboboxes
            _CB_MAP = [
                ("id_molecule", self.combobox_molecule, lambda: self._ids_molecule),
                ("id_system", self.combobox_system, lambda: self._ids_system),
                ("id_pfantmol", self.combobox_pfantmol, lambda: self._ids_pfantmol),
                ("id_statel", self.combobox_statel, lambda: self._ids_state),
                ("id_state2l", self.combobox_state2l, lambda: self._ids_state),
            ]

            for fieldname, cb, f_ids in _CB_MAP:
                idx = cb.currentIndex()
                if idx > 0:
                    self._molconsts[fieldname] = f_ids()[idx-1]


            fieldname = None
            # Values that must be in sync with some id's
            id_ = self._molconsts["id_molecule"]
            if id_ is not None:
                row = self._moldb.query_molecule(id=id_).fetchone()
                self._molconsts["formula"] = row["formula"]
                self._molconsts["name"] = row["name"]

            id_ = self._molconsts["id_pfantmol"]
            if id_ is not None:
                row = self._moldb.query_pfantmol(id=id_).fetchone()
                self._molconsts["description"] = row["description"]


        except Exception as E:
            flag_error = True
            if fieldname is not None:
                emsg = "Field '{}': {}".format(fieldname, str(E))
            else:
                emsg = str(E)
            self.add_log_error(emsg, E=E)


        self._flag_valid = not flag_error
        if not flag_error:
            self.status("")




import a99
import tabulate
import f311


__all__ = ["FileHitranDB", "populate_hitrandb"]

#
# __fileobj = None
# def get_conn():
#     return a99.get_conn(__ALIAS)


class FileHitranDB(f311.FileSQLiteDB):
    """HITRAN Molecules Catalogue"""

    default_filename = "hitrandb.sqlite"

    def _create_schema(self):
        conn = self.get_conn()
        c = conn.cursor()
        c.execute("""create table molecule (ID integer unique, Formula text unique, Name text)""")
        # **Note** isotopologue.ID is not unique, it starts over for each molecule
        c.execute("""create table isotopologue (ID integer,
                                                ID_molecule integer,
                                                Formula text unique,
                                                AFGL_Code integer,
                                                Abundance real
                                               )""")
        conn.commit()
        conn.close()

    def query_molecule(self, **kwargs):
        """Convenience function to query 'molecule' table

        Args:
            **kwargs: filter fieldname=value pairs to be passed to WHERE clause

        Returns: sqlite3 cursor
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select * from molecule{} order by ID""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def query_isotopologue(self, **kwargs):
        """Convenience function to query 'isotopologue' table

        Args, Returns: see query_molecule

        Example:

        >>> f = FileHitranDB()
        >>> f.init_default()
        >>> _ = f.query_isotopologue(**{"molecule.formula": "OH"})
        """
        where = ""
        if len(kwargs) > 0:
            where = " where " + " and ".join([key + " = ?" for key in kwargs])
        conn = self.get_conn()
        sql = """select molecule.formula as m_formula, isotopologue.* from isotopologue
                 join molecule on isotopologue.id_molecule = molecule.id{}""".format(where)
        r = conn.execute(sql, list(kwargs.values()))
        return r

    def print_isotopologues(self, **kwargs):
        """
        Prints isotopologues table in console

        Args:
            **kwargs: arguments passed to query_state()

        Example:

        >>> f = FileHitranDB()
        >>> f.init_default()
        >>> f.print_isotopologues(**{"molecule.formula": "CO"})
        """
        r = self.query_isotopologue(**kwargs)
        data, header = a99.cursor_to_data_header(r)
        print(tabulate.tabulate(data, header))


def populate_hitrandb(db):
    """Populates database with HITRAN data

    Populates a sqlite3 database represented by a FileHitranDB object with information downloaded
    from the HITRAN website

    Args:
        db: FileHitranDB instance

    .. todo:: This routine is never used
    """

    import pyfant

    assert isinstance(db, pyfant.FileHitranDB)
    conn = db.get_conn()
    try:

        mols, _ = pyfant.get_hitran_molecules()

        conn.executemany("insert into molecule values (?,?,?)", mols)

        a99.get_python_logger().info("Inserted {} molecules".format(len(mols)))

        for mol in mols:
            isos, _ = pyfant.get_hitran_isotopologues(mol[0])

            for iso in isos:
                try:
                    conn.execute("insert into isotopologue values(?,?,?,?,?)",
                                 [iso[0], mol[0]] + iso[1:])
                except:
                    a99.get_python_logger().exception("Tried to insert this: {}".format(iso))
                    raise

            a99.get_python_logger().info("Inserted {} isotopologues for molecule '{}' ({})".
                                         format(len(isos), *mol[1:3]))
    finally:
        conn.commit()
        conn.close()

    conn.close()




"""Visualization classes for pyfant file types"""

import pyfant
from f311 import Vis
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from mpl_toolkits.mplot3d import Axes3D  # yes, required (see below)
import a99


__all__ = ["VisFileToH", "draw_toh", "VisAtoms", "VisMolecules", "VisMain", "VisAbonds",]


class VisFileToH(Vis):
    """
    Plots hydrogen lines: each atmospheric layer is plotted as a y-axis-dislocated
    Spectrum in a 3D plot.
    """

    input_classes = (pyfant.FileToH,)
    action = "Visualize hydrogen lines profiles"

    def _do_use(self, r):
        fig = plt.figure()
        mpl.rcParams['legend.fontsize'] = 10
        fig.canvas.set_window_title(self.title)  # requires the Axes3D module
        # requires the Axes3D module
        ax = fig.gca(projection='3d')
        draw_toh(r, ax)
        # ax.set_zlabel('log10(Intensity)')
        # ax.set_zlabel('?')
        plt.tight_layout()
        plt.show()


def draw_toh(r, ax):
    x = np.concatenate((2 * r.lambdh[0] - r.lambdh[-2::-1], r.lambdh))
    _y = np.ones(len(x))
    for i in range(r.th.shape[1]):
        z = np.concatenate((r.th[-2::-1, i], r.th[:, i]))
        # ax.plot(x, _y * (i + 1), np.log10(z), label='a', color='k')
        ax.plot(x, _y * (i + 1), z, label='a', color='k')
    ax.set_xlabel('Wavelength ($\AA$)')
    ax.set_ylabel("Atmospheric layer #")


# # Editors

class VisAtoms(Vis):
    """Opens the ated window."""
    input_classes = (pyfant.FileAtoms,)
    action = "Edit using atomic lines editor"

    def _do_use(self, r):
        form = a99.keep_ref(pyfant.XFileAtoms(self.parent_form))
        form.load(r)
        form.show()


class VisMolecules(Vis):
    """Opens the mled window."""
    input_classes = (pyfant.FileMolecules,)
    action = "Edit using molecular lines editor"

    def _do_use(self, r):
        form = a99.keep_ref(pyfant.XFileMolecules(self.parent_form))
        form.load(r)
        form.show()


class VisMain(Vis):
    """Opens the mained window."""
    input_classes = (pyfant.FileMain,)
    action = "Edit using main configuration file editor"

    def _do_use(self, r):
        form = a99.keep_ref(pyfant.XFileMain(self.parent_form, r))
        form.show()


class VisAbonds(Vis):
    """Opens the abed window."""
    input_classes = (pyfant.FileAbonds,)
    action = "Edit using abundances file editor"

    def _do_use(self, r):
        form = a99.keep_ref(pyfant.XFileAbonds(self.parent_form, r))
        form.show()


class VisWhatever(Vis):
    cls_form = None

    def _do_use(self, r):
        cls_form = self._get_cls_form()
        form = a99.keep_ref(cls_form(self.parent_form, r))
        form.show()

    def _get_cls_form(self):
        """Method to return the class of the form to instantiate"""
        raise NotImplementedError()


class VisOptions(VisWhatever):
    """Allows for editing a FileOptions object"""
    input_classes = (pyfant.FileOptions,)
    action = "Edit using command-line options file editor"

    def _get_cls_form(self):
        return pyfant.XFileOptions




#!/usr/bin/env python

"""
Editor for molecular constants file

This application can edit files of class FileMolConsts.
"""

import sys
import argparse
import a99
import logging
import os
import pyfant


a99.logging_level = logging.INFO
a99.flag_log_file = True


if __name__ == "__main__":

    deffn = pyfant.FileMolConsts.default_filename

    parser = argparse.ArgumentParser(
    description=__doc__,
    formatter_class=a99.SmartFormatter
    )
    parser.add_argument('fn', type=str, help='Molecular constants file name',
                        default=deffn, nargs='?')
    args = parser.parse_args()

    if args.fn == deffn and not os.path.isfile(deffn):
        args.fn = None

    fobj = None
    if args.fn is not None:
        fobj = pyfant.FileMolConsts()
        fobj.load(args.fn)
    app = a99.get_QApplication([])
    form = pyfant.XFileMolConsts(None)
    form.load(fobj)
    form.show()

    sys.exit(app.exec_())










# Conversion from hapi-loaded data was discontinued: conv_hitran160.py now

# """Molecular lines converter from HITRAN to PFANT format."""
#
# __all__ = ["ConvHITRAN"]
#
# import a99, airvacuumvald as avv
# from . import Conv
#
# class ConvHITRAN(Conv):
#     """Converts HITRAN molecular lines data to PFANT "sets of lines" using Einstein's coefficients
#
#     Args:
#         strengthfactor: (=1.) all resulting line strengths ("sj") will be multiplied by this factor.
#                         Note: it is better to use "fe" argument instead
#         isowant: (=1) isotopologue ID according to HITRAN or literature e.g. [1]. Li et al. 2015 didn't follow
#                  HITRAN's isotopologues codes for CO
#
#     This conversor expects data in hapi format. The hapi library is provided by HITRAN itself. The way it works is it
#     parses all ".par" files in a given directory before you can choose which one to use. Therefore, it may be a good
#     idea to isolate the file of interest or create link to file.
#
#     Both strengthfactor and fe are ways to achieve the same result (multiply the line strength), but the former applies
#     to sj whereas the latter is recorded as the molecule's "FE"
#
#     There is a working example conv_hitran_co.py in examples directory
#
#     References:
#         [1] Li, G., Gordon, I.E., Rothman, L.S., Tan, Y., Hu, S.M., Kassi, S., Campargue, A. and Medvedev, E.S., 2015.
#         Rovibrational line lists for nine isotopologues of the CO molecule in the X1Σ+ ground electronic state.
#         The Astrophysical Journal Supplement Series, 216(1), p.15.
#
#         [2] Plez's conversor translate_molec_linelists_v16.1.f
#     """
#
#     def __init__(self, *args, isowant=1, strengthfactor=1., **kwargs):
#         super().__init__(*args, **kwargs)
#         if not (1 <= isowant <= 9):
#             raise ValueError(f"Invalid 'isowant': {isowant} (must be between 1 and 9)")
#         self.isowant = isowant
#         self.strengthfactor = strengthfactor
#
#     def _make_sols(self, hapidata):
#         """Sets-of-lines maker for ConvHITRAN class
#
#         Args:
#             hapidata: this is hapi.LOCAL_TABLE_CACHE["some_name"]
#         """
#         import pyfant
#
#         log = self.log
#         sols = self.sols
#
#         hapidata_ = hapidata["data"]
#         n = log.n = len(hapidata_["molec_id"])
#         if n == 0:
#             raise RuntimeError("Zero lines found")
#
#         local_iso_id = hapidata_["local_iso_id"]
#         s_vl = hapidata_["global_upper_quanta"]
#         s_v2l = hapidata_["global_lower_quanta"]
#         gp = hapidata_["gp"]
#         gpp = hapidata_["gpp"]
#         a = hapidata_["a"]  # Einstein's coefficient in unit 1/s
#         nu = hapidata_["nu"]  # Wavenumber in unit 1/cm
#
#         deltak = self.molconsts.get_deltak()
#         S2l = self.molconsts.get_S2l()
#
#         for i in range(n):
#             try:
#                 iso = local_iso_id[i]
#                 if iso == self.isowant:
#                     log.cnt_in += 1
#                 else:
#                     log.skip_reasons[f"isotopologue {iso}"] += 1
#                     continue
#
#                 vl = int(s_vl[i])
#                 v2l = int(s_v2l[i])
#                 Jl = (gp[i]-1.)/2.  # Taken from Plez's conversor [2]
#                 J2l = (gpp[i]-1.)/2.
#
#                 # normalizationfactor = self.strengthfactor/((2*S2l+1)*(2*J2l+1)*(2-deltak))
#                 #
#                 # SJ = normalizationfactor*a[i]*1.499*(2+Jl+1)/nu[i]**2
#                 SJ = self.get_sj_einstein(a[i], Jl, J2l, S2l, deltak, nu[i], self.strengthfactor)
#
#                 lambda_ = avv.vacuum_to_air(1e8/nu[i])
#
#             except Exception as e:
#                 reason = a99.str_exc(e)
#                 log.skip_reasons[reason] += 1
#                 msg = "#{}{} line: {}".format(i+1, a99.ordinal_suffix(i+1), reason)
#                 log.errors.append(msg)
#                 if not self.flag_quiet:
#                     a99.get_python_logger().exception(msg)
#                 continue
#
#             # TODO skipping the branch altogether: PFANT does not care really
#             sols.append_line2(vl, v2l, lambda_, SJ, J2l, "-")
#
#         return sols, log
#
#
# # === SAMPLES ==========================================================================================================
# # hapidata["header"]
# # ==============
# # {'table_type': 'column-fixed',
# #  'size_in_bytes': -1,
# #  'table_name': '###',
# #  'number_of_rows': 253695,
# #  'order': ['molec_id',
# #   'local_iso_id',
# #   'nu',
# #   'sw',
# #   'a',
# #   'gamma_air',
# #   'gamma_self',
# #   'elower',
# #   'n_air',
# #   'delta_air',
# #   'global_upper_quanta',
# #   'global_lower_quanta',
# #   'local_upper_quanta',
# #   'local_lower_quanta',
# #   'ierr',
# #   'iref',
# #   'line_mixing_flag',
# #   'gp',
# #   'gpp'],
# #  'format': {'a': '%10.3E',
# #   'gamma_air': '%5.4f',
# #   'gp': '%7.1f',
# #   'local_iso_id': '%1d',
# #   'molec_id': '%2d',
# #   'sw': '%10.3E',
# #   'local_lower_quanta': '%15s',
# #   'local_upper_quanta': '%15s',
# #   'gpp': '%7.1f',
# #   'elower': '%10.4f',
# #   'n_air': '%4.2f',
# #   'delta_air': '%8.6f',
# #   'global_upper_quanta': '%15s',
# #   'iref': '%12s',
# #   'line_mixing_flag': '%1s',
# #   'ierr': '%6s',
# #   'nu': '%12.6f',
# #   'gamma_self': '%5.3f',
# #   'global_lower_quanta': '%15s'},
# #  'default': {'a': 0.0,
# #   'gamma_air': 0.0,
# #   'gp': 'FFF',
# #   'local_iso_id': 0,
# #   'molec_id': 0,
# #   'sw': 0.0,
# #   'local_lower_quanta': '000',
# #   'local_upper_quanta': '000',
# #   'gpp': 'FFF',
# #   'elower': 0.0,
# #   'n_air': 0.0,
# #   'delta_air': 0.0,
# #   'global_upper_quanta': '000',
# #   'iref': 'EEE',
# #   'line_mixing_flag': 'EEE',
# #   'ierr': 'EEE',
# #   'nu': 0.0,
# #   'gamma_self': 0.0,
# #   'global_lower_quanta': '000'},
# #  'description': {'a': 'Einstein A-coefficient in s-1',
# #   'gamma_air': 'Air-broadened Lorentzian half-width at half-maximum at p = 1 atm and T = 296 K',
# #   'gp': 'Upper state degeneracy',
# #   'local_iso_id': 'Integer ID of a particular Isotopologue, unique only to a given molecule, in order or abundance (1 = most abundant)',
# #   'molec_id': 'The HITRAN integer ID for this molecule in all its isotopologue forms',
# #   'sw': 'Line intensity, multiplied by isotopologue abundance, at T = 296 K',
# #   'local_lower_quanta': 'Rotational, hyperfine and other quantum numbers and labels for the lower state of a transition',
# #   'local_upper_quanta': 'Rotational, hyperfine and other quantum numbers and labels for the upper state of a transition',
# #   'gpp': 'Lower state degeneracy',
# #   'elower': 'Lower-state energy',
# #   'n_air': 'Temperature exponent for the air-broadened HWHM',
# #   'delta_air': 'Pressure shift induced by air, referred to p=1 atm',
# #   'global_upper_quanta': 'Electronic and vibrational quantum numbers and labels for the upper state of a transition',
# #   'iref': 'Ordered list of reference identifiers for transition parameters',
# #   'line_mixing_flag': 'A flag indicating the presence of additional data and code relating to line-mixing',
# #   'ierr': 'Ordered list of indices corresponding to uncertainty estimates of transition parameters',
# #   'nu': 'Transition wavenumber',
# #   'gamma_self': 'Self-broadened HWHM at 1 atm pressure and 296 K',
# #   'global_lower_quanta': 'Electronic and vibrational quantum numbers and labels for the lower state of a transition'}}
#
# # Sample lines, although I won't have to parse them because hapi does that
# # ========================================================================
# # 55    2.248764 2.700-164 8.704E-07.07970.08664763.01990.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# # 55    2.279785 6.503-162 8.685E-07.07970.08663631.47680.76-.000265             40             40                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# # 56    2.288013 3.667-166 9.362E-07.07970.08665317.82150.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7    36.0   12.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# # 55    2.310683 1.737-159 8.638E-07.07970.08662478.03720.76-.000265             39             39                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# # 56    2.320259 8.961-164 9.357E-07.07970.08664183.55380.76-.000265             40             40                    R  0      457665 5 8 2 2 1 7    36.0   12.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# # 53    2.325385 1.473-164 1.002E-06.07970.08665843.01970.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7     3.0    1.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
# # 52    2.331616 5.471-164 1.014E-06.07970.08665930.07650.76-.000265             41             41                    R  0      457665 5 8 2 2 1 7     6.0    2.0 0.0778 0.671 0.000180 0.1037 0.780 -.000800
#
#
# #=======================================================================================================================
# # (20230503) Old code from 2017, didn't delete yet, maybe todo cleanup
# # """
# # HITRAN-specific conversion
# #
# # References:
# #   [1] Rothman, Laurence S., et al. "The HITRAN 2004 molecular spectroscopic database."
# #       Journal of Quantitative Spectroscopy and Radiative Transfer 96.2 (2005): 139-204.
# #   [2] extraehitran.f (First version: Jorge Melendez-Moreno, January 1999)
# # """
# #
# # # TODO not working
# #
# #
# # from .convlog import *
# # from collections import OrderedDict
# # import airvacuumvald as avv
# # import pyfant
# #
# #
# # __all__ = ["hitran_to_sols"]
# #
# #
# # _OH_BRANCH_MAP_OH = {"QQ": "Q", "PP": "P", "RR": "R"}
# #
# #
# #
# # def _parse_local_lower_quanta_OH(Q2l):
# #     """
# #     Parses the "Lower-state 'local' quanta" for 'OH' (see [1], Table 4, group 2, observation 'b')
# #
# #     Args:
# #         Q2l: 15-character string containing the lower-state 'local' quanta information of molecular
# #         line. Example: " PP 12.5ff     "
# #
# #     Returns: (Br, J2l), where
# #         Br: branch ("P"/"Q"/"R")
# #         J2l: inferior "quantum number associated with the total angular momentum excluding nuclear
# #              spin" [1]
# #
# #     **Note** will raise if cannot determine Br (branch)
# #     """
# #     brinfo = Q2l[1:3]
# #     Br = _OH_BRANCH_MAP_OH.get(brinfo)
# #     if Br is None:
# #         raise ValueError("Cannot derive branch from '{}' ((local lower quanta) = '{}')".
# #                          format(brinfo, Q2l))
# #     J2l = float(Q2l[3:8])
# #     return Br, J2l
# #
# # def _parse_local_lower_quanta_group2(Q2l):
# #     """
# #     Parses the "lower-state 'local' quanta" for molecules in [1], Table 4, group 2
# #
# #     Args, Returns: see _parse_local_lower_quanta_OH()
# #     """
# #
# #     Br = Q2l[5:6]
# #     J2l = float(Q2l[6:9])
# #     return Br, J2l
# #
# #
# # def _parse_local_lower_quanta_group6(Q2l):
# #     """[...] [1], Table 4, group 6"""
# #
# #     Br = Q2l[3:4]
# #     J2l = float(Q2l[4:9])
# #     return Br, J2l
# #
# #
# # def _parse_global_quanta_class1(V):
# #     """
# #     Parses the [upper/lower] "global quanta" ([1], Table 3, class 1
# #
# #     Args:
# #         V: 15-character string containing the 'global' quanta information of molecular
# #            line. Example: "       X3/2   8"
# #
# #     Returns: v1: "quantum number associated with the first normal mode of vibration" [1]
# #     """
# #     return int(V[13:15])
# #
# #
# # def _parse_global_quanta_class2(V):
# #     """[...] [1], Table 3, class 2"""
# #     return int(V[13:15])
# #
# # def _parse_global_quanta_class3(V):
# #     return int(V[13:15])
# #
# #
# # __GROUP_MAP = (
# #     (("OH",),
# #      _parse_local_lower_quanta_OH),
# #     (("CO2", "N2O", "CO", "HF", "HCl", "HBr", "HI", "OCS", "N2", "HCN", "C2H2", "NO+"),
# #      _parse_local_lower_quanta_group2),
# #     (("NO", "ClO"),
# #      _parse_local_lower_quanta_group6),
# # )
# # _GROUP_DICT = dict(sum([[(mol, callable_) for mol in mols] for mols, callable_ in __GROUP_MAP], []))
# #
# # __CLASS_MAP = (
# #     (("CO", "HF", "HCl", "HBr", "HI", "N2", "NO+"), _parse_global_quanta_class1),
# #     (("O2",), _parse_global_quanta_class2),
# #     (("NO", "OH", "ClO"), _parse_global_quanta_class3),
# # )
# # _CLASS_DICT = dict(sum([[(mol, callable_) for mol in mols] for mols, callable_ in __CLASS_MAP], []))
# #
# #
# # def hitran_to_sols(molconsts, lines, qgbd_calculator):
# #     """
# #     Converts HITRAN molecular lines data to PFANT "sets of lines"
# #
# #     Args:
# #         molconsts: a dict-like object combining field values from tables 'molecule', 'state',
# #                     'pfantmol', and 'system' from a FileMolDB database
# #         lines: item from rom hapi.LOCAL_TABLE_CACHE (a dictionary)
# #         qgbd_calculator: callable that can calculate "qv", "gv", "bv", "dv",
# #                          e.g., calc_qbdg_tio_like()
# #
# #     Returns: (a list of ftpyfant.SetOfLines objects, a MolConversionLog object)
# #     """
# #
# #     def append_error(msg):
# #         log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))
# #
# #     # C     BAND (v',v'')=(VL,V2L)
# #     # C     WN: vacuum wavenumber   WL : air wavelength
# #     # C     J2L: lower state angular momentum number
# #     # C     iso: isotope/ 26: 12C16O, 36: 13C16O, 27: 12C17O, 28: 12C18O
# #     # C     ramo: branch, for P: ID=1, for R: ID=2
# #     # C     name: file name/ coisovLv2L
# #     # C     HL: Honl-London factor
# #     # C     FR: oscillator strength
# #
# #     header = lines["header"]
# #     data = lines["data"]
# #
# #     S = molconsts.get_S2l()
# #     DELTAK = molconsts.get_deltak()
# #     print(f"DELTAK = {DELTAK}")
# #     sys.exit()
# #     formula = molconsts["formula"]
# #
# #     n = header["number_of_rows"]
# #     log = MolConversionLog(n)
# #     sols = OrderedDict()  # one item per (vl, v2l) pair
# #
# #     try:
# #         f_class = _CLASS_DICT[formula]
# #     except KeyError:
# #         log.errors.append("Cannot determine Table 3 'class' (Rothman et al. 2005) for formula '{}'".
# #                           format(formula))
# #         log.flag_ok = False
# #
# #     try:
# #         f_group = _GROUP_DICT[formula]
# #     except KeyError:
# #         log.errors.append("Cannot determine Table 4 'group' (Rothman et al. 2005) for formula '{}'".
# #                           format(formula))
# #         log.flag_ok = False
# #
# #     if not log.flag_ok:
# #         return [], log
# #
# #     for i in range(n):
# #
# #         try:
# #             nu = data["nu"][i]
# #             wl = avv.vacuum_to_air(1e8/nu)
# #             Br, J2l = f_group(data["local_lower_quanta"][i])
# #             #  ASSUMING SINGLET!!!
# #             Jl = ph.singlet.quanta_to_branch(Br, J2l)
# #             V = f_class(data["global_upper_quanta"][i])
# #             V_ = f_class(data["global_lower_quanta"][i])
# #             A = data["a"][i]
# #
# #             # A seguir normalizacion de factor de Honl-london (HLN)
# #             Normaliza = 1/((2.0*J2l+1)*(2.0*S+1)*(2.0-DELTAK))
# #
# #             # A seguir teremos a forca de oscilador
# #             # mas q sera normalizada segundo o programa da Beatriz
# #             gf = Normaliza*1.499*(2*Jl+1)*A/(nu**2)
# #
# #             J2l_pfant = int(J2l)  # ojo, estamos colocando J2L-0.5! TODO ask BLB: we write J2l or J2l-.5? (20171114 I think this is wrong, molecules.dat has decimal values)
# #         except Exception as e:
# #             log.errors.append("#{}{} line: {}".format(i+1, a99.ordinal_suffix(i+1), a99.str_exc(e)))
# #             continue
# #
# #         sol_key = (V, V_)  # (v', v'') transition (v_sup, v_inf)
# #         if sol_key not in sols:
# #             qgbd = qgbd_calculator(molconsts, sol_key[1])
# #             qqv = qgbd["qv"]
# #             ggv = qgbd["gv"]
# #             bbv = qgbd["bv"]
# #             ddv = qgbd["dv"]
# #
# #             sols[sol_key] = pyfant.SetOfLines(V, V_, qqv, ggv, bbv, ddv, 1.)
# #         sol = sols[sol_key]
# #         sol.append_line(wl, gf, J2l_pfant, Br)
# #
# #     return (list(sols.values()), log)



"""Widget to edit a MargAbondsFwhm object."""


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import sys
import types
import traceback
import a99
import pyfant


__all__ = ["WFileAbXFwhm"]


class WFileAbXFwhm(QWidget):
    """
    Editor for editor widget.

    Args:
      parent=None
    """

    # Emitted whenever any value changes
    changed = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        # # Setup & accessible attributes

        # Whether all the values in the fields are valid or not
        self.flag_valid = False  # initialized to False because not loaded yet
        self.f = None  # FileOptions object
        self.logger = a99.get_python_logger()
        # FileAbonds instance to check for existence of atomic symbols
        self.file_abonds = None


        # # Internal stuff that must not be accessed from outside

        # options map: list of _Option
        self.omap = []
        # Internal flag to prevent taking action when some field is updated programatically
        self.flag_process_changes = False

        # # Central layout

        la = self.layout_main = QVBoxLayout()
        a99.set_margin(la, 0)
        self.setLayout(la)

        # ## Splitter with scroll area and descripton+error area
        sp = self.splitter = QSplitter(Qt.Vertical)
        la.addWidget(sp)

        # ### Text edit

        # ### Editor for multi setup
        editor = self.editor = QPlainTextEdit()
        # editor.setStyleSheet("QPlainTextEdit {background-color: #000000}")
        # editor.textChanged.connect(self.on_edited)
        editor.modificationChanged.connect(self.on_edited)
        sp.addWidget(editor)
        self.highlight = a99.PythonHighlighter(editor.document())


        # ### Second widget of splitter
        # layout containing description area and a error label
        wlu = QWidget()
        lu = QVBoxLayout(wlu)
        a99.set_margin(lu, 0)
        lu.setSpacing(4)
        x = self.textEditInfo = QTextEdit(self)
        x.setReadOnly(True)
        x.setStyleSheet("QTextEdit {color: %s}" % a99.COLOR_DESCR)
        lu.addWidget(x)
        sp.addWidget(wlu)

        # ### Adjust splitter proportion
        sp.setStretchFactor(0, 8)
        sp.setStretchFactor(1, 2)

        self.setEnabled(False)  # disabled until load() is called
        a99.style_checkboxes(self)
        self.flag_process_changes = True

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Interface

    def load(self, x):
        assert isinstance(x, pyfant.FileAbXFwhm)
        self.f = x
        self.__update_from_data()
        # this is called to perform file validation upon loading
        self.validate()
        
        self.setEnabled(True)

    def validate(self):
        """Forces validation of text in editor."""
        if self.f:
            self.__update_data()
        else:
            self.flag_valid = False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Qt override

    def setFocus(self, reason=None):
        pass  # self.editor.setFocus()

#    def eventFilter(self, obj_focused, event):
#        pass
#        return False

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Slots

    def on_edited(self):
        # http://stackoverflow.com/questions/22685463/qt5-tell-qplaintextedit-to-ignore-syntax-highlighting-changes
        if not self.flag_process_changes:
            return
        self.editor.document().setModified(False)
        self.__update_data()
        self.changed.emit()

    # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * # * #
    # # Internal gear

    def __validate(self):
        f = pyfant.FileAbXFwhm()
        # first validation: code parses OK and has "ab" and "conv" variables
        f.source = str(self.editor.toPlainText())
        f.validate(self.file_abonds)

    def __update_from_data(self):
        self.flag_process_changes = False
        try:
            self.editor.setPlainText(self.f.source)
        finally:
            self.flag_process_changes = True

    def __update_data(self):
        emsg, flag_error = "", False
        try:
            self.__validate()
            self.f.source = str(self.editor.toPlainText())

        except:
            flag_error = True
            etype, value, _ = sys.exc_info()
            emsg = "<b>Error</b><br><pre>"+\
                   ("".join(_format_exception_only(etype, value)))+"</pre>"
            # ShowError(str(E))
        self.flag_valid = not flag_error
        self.__set_descr_text('<span style="color: %s">%s</div>' % (a99.COLOR_ERROR, emsg))

    def __set_descr_text(self, x):
        """Sets text of labelDescr."""
        self.textEditInfo.setText(x)


def _format_exception_only(etype, value):
    """[Copied from the traceback module] Format the exception part of a traceback.

    [Copied from the traceback module and changed to ignore filename]

    The arguments are the exception type and value such as given by
    sys.last_type and sys.last_value. The return value is a list of
    strings, each ending in a newline.

    Normally, the list contains a single string; however, for
    SyntaxError exceptions, it contains several lines that (when
    printed) display detailed information about where the syntax
    error occurred.

    The message indicating which exception occurred is always the last
    string in the list.

    """

    # An instance should not have a meaningful value parameter, but
    # sometimes does, particularly for string exceptions, such as
    # >>> raise string1, string2  # deprecated
    #
    # Clear these out first because issubtype(string1, SyntaxError)
    # would raise another exception and mask the original problem.
    if (isinstance(etype, BaseException) or
        isinstance(etype, object) or
        etype is None or type(etype) is str):
        return [traceback._format_final_exc_line(etype, value)]

    stype = etype.__name__

    if not issubclass(etype, SyntaxError):
        return [_format_final_exc_line(stype, value)]

    # It was a syntax error; show exactly where the problem was found.
    lines = []
    try:
        msg, (filename, lineno, offset, badline) = value.args
    except Exception:
        pass
    else:
        lines.append('Line %d\n' % lineno)
        if badline is not None:
            lines.append('    %s\n' % badline.strip())
            if offset is not None:
                caretspace = badline.rstrip('\n')
                offset = min(len(caretspace), offset) - 1
                caretspace = caretspace[:offset].lstrip()
                # non-space whitespace (likes tabs) must be kept for alignment
                caretspace = ((c.isspace() and c or ' ') for c in caretspace)
                lines.append('    %s^\n' % ''.join(caretspace))
        value = msg

    lines.append(traceback._format_final_exc_line(stype, value))
    return lines


def _format_final_exc_line(etype, value):
    """[Copied from traceback module] Return a list of a single line -- normal case for format_exception_only.

    [Changed to ignore the exception type]"""
    valuestr = traceback._some_str(value)
    if value is None or not valuestr:
        line = "%s\n" % etype
    else:
        line = "%s\n" % valuestr
    return line



from .convlog import *
from .conv import *
import pyfant

__all__ = ["ConvPlez"]


class ConvPlez(Conv):
    """
    Converts Plez molecular lines data to PFANT "sets of lines"

        Args:
            species: species name within Plez's file. Examples: "NH A-X PGopher", "Sneden web" etc.

    ***Note** old format (FilePlezMoleculeOld) does not have loggf information,
              therefore the only way to work with the latter file type is
              mode==CONVPLEZMODE_HLF
    """

    def __init__(self, *args, species="", **kwargs):
        Conv.__init__(self, *args, **kwargs)
        self.species = species

    def _make_sols(self, file):

        log = self.log
        sols = self.sols

        def append_error(msg):
            log.errors.append("#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), str(msg)))

        if not isinstance(file, pyfant.FilePlezLinelistBase):
            raise TypeError("Invalid type for argument 'fileobj': {}".format(type(file).__name__))

        try:
            lines = file.molecules[self.species].lines
        except KeyError:
            raise KeyError(f"Species '{self.species}' not found in {list(file.molecules.keys())}")
        n = log.n = len(lines)

        if n == 0:
            raise RuntimeError("Species '{}' has zero lines".format(self.species))

        if self.mode == ConvMode.HLF:
            mtools = self.kovacs_toolbox()
        else:
            deltak = self.molconsts.get_deltak()
            S2l = self.molconsts.get_S2l()

        if self.flag_filter_labels:
            if self.mode == ConvMode.EINSTEIN_MINIMAL:
                raise ValueError(f"mode==ConvMode.EINSTEIN_MINIMAL and flag_filter_label==True are incompatible")

            from_label = self.molconsts["from_label"]
            to_label = self.molconsts["to_label"]

        for i, line in enumerate(lines):
            branch = line.branch
            sj = 1.

            try:
                if self.flag_filter_labels:
                    if line.from_label != from_label or line.to_label != to_label:
                        log.skip_reasons[f"Wrong system: {line.from_label}-{line.to_label}"] += 1
                        continue

                if self.mode == ConvMode.HLF:
                    try:
                        hlf = mtools.get_sj(line.vl, line.v2l, line.J2l, branch)
                    except pyfant.NoLineStrength:
                        log.skip_reasons["Cannot calculate HLF"] += 1
                        continue

                    if hlf < 0:
                        log.skip_reasons["Negative SJ"] += 1
                        continue

                    sj *= hlf

                else:
                    if self.mode == ConvMode.EINSTEIN_MINIMAL:
                        J = (line.gu-1.)/2.
                        if J < 0: J = 0
                        line.Jl = line.J2l = J
                        line.vl = line.v2l = 0

                    if not branch:
                        branch = "-"

                    normalizationfactor = self.get_normalizationfactor(line.J2l, S2l, deltak)
                    sj *= normalizationfactor*10**line.loggf

                self.append_line(line, sj, branch)

            except Exception as e:
                reason = a99.str_exc(e)
                log.skip_reasons[reason] += 1
                msg = "#{}{} line: {}".format(i + 1, a99.ordinal_suffix(i + 1), reason)
                log.errors.append(msg)
                if not self.flag_quiet:
                    a99.get_python_logger().exception(msg)
                continue

        return sols, log
